"""
This module contains the FilocIO class, used to work with files defined by a locpath: a path containing format placeholders.
"""
import logging
import os
import re
import uuid
from io import UnsupportedOperation
from typing import Dict, Any, List, Mapping, Optional, Set
from typing import Tuple

import fsspec
import parse
from fsspec import AbstractFileSystem
from fsspec.core import OpenFile

from filoc.contract import Constraints, Constraint

log = logging.getLogger('filoc')

# ---------
# Constants
# ---------
_re_natural           = re.compile(r"(\d+)")
_re_path_placeholder  = re.compile(r'({[^}]+})')


# -------
# Helpers
# -------
def sort_natural(li: List[str]) -> List[str]:
    """ Perform natural sort of string containing numbers. Floating number are currently supported but cannot be compared to integers (missing dot separator)"""
    # TODO: support mix of int and float
    return sorted(li, key=lambda s: [int(part) if part.isdigit() else part for part in _re_natural.split(s)])


def coerce_nullable_mapping(d) -> Optional[Mapping[str, Any]]:
    """
    Pass through Mapping or None instance, tries to call ``to_dict()`` method elsewhere
    Args:
        d:

    Returns:
        the coerced mapping of type ``Mapping[str, Any]`` or None
    Raises:
        TypeError: if ``d`` is neither an instance of Mapping, nor None, nor contains `to_dict()` method
    """
    if d is None:
        return d
    if isinstance(d, Mapping):
        return d
    if getattr(d, "to_dict", None):
        # especially valid for pandas Series
        return d.to_dict()
    raise TypeError(f"Expected instance of Mapping or implementing to_dict, got {type(d)}!")


def mix_dicts_and_coerce(dict1, dict2) -> Mapping[str, Any]:
    """
    Coerce dict1 and dict2 to Mapping and mix both together
    Args:
        dict1: first dictionary argument (either Mapping or implements `to_dict()`)
        dict2: second dictionary argument (either Mapping or implements `to_dict()`)

    Returns:
        The combined Mapping
    """
    dict1 = coerce_nullable_mapping(dict1)
    dict2 = coerce_nullable_mapping(dict2)

    dict2 = None if len(dict2) == 0 else dict2
    if dict1 and dict2:
        result = dict()
        result.update(dict1)
        result.update(dict2)
        return result
    elif dict1:
        return dict1
    elif dict2:
        return dict2
    else:
        return dict()


# TODO: Support path character escaping


# -------------------
# Class FilocIO
# -------------------
class FilocIO:
    """
    Class providing access to files, given the locpath definition. The locpath is a format string, which placeholders
    are variables used by FilocIO to map variables to paths and paths to variables.
    """
    def __init__(
        self, locpath: str, 
        writable: bool = False, 
        fs: AbstractFileSystem = None
    ) -> None:
        super().__init__()
        self._original_locpath = locpath
        self._writable  = writable

        # split locpath to distinguish placeholders from constant parts
        path_elts = _re_path_placeholder.split(locpath)

        # Normalize the input path, by creating an fsspec OpenFile, then by getting the path property, 
        # which is normalized. But the placeholders within the locpath are not valid, so we replace them by
        # a valid random string, build the OpenFile, get the normalized string, and replace the random
        # string by the original placeholders.
        path_elts_and_ersatz = [ (elt, str(uuid.uuid4()) if elt.startswith("{") else None) for elt in path_elts ]
        some_valid_path = "".join([ersatz if ersatz else elt for elt, ersatz in path_elts_and_ersatz])
        if fs is None:
            open_file = fsspec.open(some_valid_path)
        else:
            open_file = OpenFile(fs, some_valid_path)

        # now build the normalized locpath, by replacing ersatz string by the original placeholder strings
        self._locpath = open_file.path
        for elt, ersatz in path_elts_and_ersatz:
            if ersatz:
                self._locpath = self._locpath.replace(ersatz, elt)

        self._fs = open_file.fs  # type: AbstractFileSystem
        self._path_parser = parse.compile(self._locpath)  # type: parse.Parser

        # Get the root folder: the last folder, that is not variable
        self._root_folder = self._locpath.split("{")[0] 
        self._root_folder = self.fs.sep.join((self._root_folder + "dummy_to_ensure_subfolder").split(self.fs.sep)[:-1])  

        # parse library contains the _named_fields property, which provides us with the set of placeholder names
        # noinspection PyProtectedMember
        self._path_props  = set(self._path_parser._named_fields)  # type: Set[str]

    @property
    def fs(self) -> AbstractFileSystem:
        return self._fs

    # noinspection PyDefaultArgument
    def parse_path_properties(self, path: str) -> Dict[str, Any]:
        """
        Extract the ``self.locpath`` placeholder values contained in ``path``
        Args:
            path: path string

        Returns:
            A dictionary containing the "placeholder name" -> value mapping
        """
        try:
            return self._path_parser.parse(path).named
        except Exception as e:
            raise ValueError(f'Could not parse {path} with {self._locpath} parser: {e}')

    def render_path(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> str:
        """
        Render the path defined by the provided placeholder values (``constraints``).
        Args:
            constraints: The placeholders values required by ``locpath``.
            **constraints_kwargs: The placeholders values required by ``locpath``.

        Returns:
            The rendered path
        Raises:
            ValueError: If a placeholder value is missing
        """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        undefined_keys = self._path_props - set(constraints)

        if len(undefined_keys) > 0:
            raise ValueError('Required props undefined: {}. Provided: {}'.format(undefined_keys, constraints))
        return self._locpath.format(**constraints)  # result should be normalized, because locpath is

    def render_glob_path(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> str:
        """
        Render a glob path defined by the provided placeholder values (``constraints``). The missing missing placeholders
        are replaced by ``*`` in the glob path.
        Args:
            constraints: The placeholders values defined in ``locpath``.
            **constraints_kwargs: The placeholders values defined in ``locpath``.

        Returns:
            A glob path
        """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        provided_keys = set(constraints)

        undefined_keys = self._path_props - provided_keys
        defined_keys = self._path_props - undefined_keys

        path_values = dict()
        path_values.update({(k, constraints[k]) for k in defined_keys})

        glob_path = self._locpath
        for undefined_key in undefined_keys:
            glob_path = re.sub(r'{' + undefined_key + r'(?::[^}]*)?}', '?*', glob_path)

        # finally format
        glob_path = glob_path.format(**path_values)
        return glob_path  # result should be normalized, because locpath is

    def list_paths(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> List[str]:
        """
        Gets the list of all existing and valid paths fulfilling the provided constraints
        Args:
            constraints: The equality constraints applied to the ``locpath`` placeholders
            **constraints_kwargs: The equality constraints applied to the ``locpath`` placeholders

        Returns:
            The list of valid and existing paths fulfilling the provided constraints
        """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        paths = self.fs.glob(self.render_glob_path(constraints))
        return sort_natural(paths)

    def list_paths_and_props(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> List[Tuple[str, Dict[str, Any]]]:
        """
        Gets the list of all existing and valid paths fulfilling the provided constraints, along with the list of associated placeholder values
        Args:
            constraints: The equality constraints applied to the ``locpath`` placeholders
            **constraints_kwargs: The equality constraints applied to the ``locpath`` placeholders

        Returns:
            A list of tuples containing for each valid path, the path and the list of related placeholder values
        """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        paths = self.list_paths(constraints)
        return [(p, self.parse_path_properties(p)) for p in paths]

    def exists(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> bool:
        """
        Checks if the path defined by the provided placeholder values (``constraints``) exists
        Args:
            constraints: The equality constraints applied to the ``locpath`` placeholders
            **constraints_kwargs: The equality constraints applied to the ``locpath`` placeholders

        Returns:
            True if the path exists, False elsewhere
        """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        return self.fs.exists(self.render_path(constraints))

    def open(
            self,
            constraints   : Constraints,
            mode          : str = "rb",
            block_size    : int = None,
            cache_options : Optional[Dict] = None,
            **kwargs
    ):
        """
        Opens the path defined by the provided placeholder values (``constraints``)

        Args:
            constraints: The equality constraints applied to the ``locpath`` placeholders
            mode: See builtin ``open()``
            block_size: Some indication of buffering - this is a value in bytes
            cache_options: Extra arguments to pass through to the cache.
            **kwargs: Additional keyed arguments passed to ``fsspec.OpenFile.open(...)``

        Returns:
            a file-like object from the underlying fsspec filesystem
        """
        is_writing = len(set(mode) & set("wa+")) > 0
        if is_writing and not self._writable:
            raise UnsupportedOperation('this filoc is not writable. Set writable flag to True to enable writing')

        path = self.render_path(constraints)

        dirname = os.path.dirname(path)

        if is_writing:
            self.fs.makedirs(dirname, exist_ok=True)

        return self.fs.open(path, mode, block_size, cache_options, **kwargs)

    # noinspection PyDefaultArgument
    def delete(self, constraints : Optional[Constraints] = {}, dry_run=False):
        """
        Delete the path defined by the provided placeholder values (``constraints``)

        Args:
            constraints: The equality constraints applied to the ``locpath`` placeholders
            dry_run: If True, only simulates the deletion
        """

        # TODO: Unit test to test deletion of folders

        if not self._writable:
            raise UnsupportedOperation('this filoc is not writable. Set writable flag to True to enable deleting')

        path_to_delete = self.list_paths(constraints)

        dry_run_log_prefix = '(dry_run) ' if dry_run else ''
        log.info(f'{dry_run_log_prefix}Deleting {len(path_to_delete)} files with path_props "{constraints}"')
        for path in path_to_delete:
            log.info(f'{dry_run_log_prefix}Deleting "{path}"')
            if dry_run:
                continue
            if self.fs.isfile(path):
                self.fs.delete(path)
            elif self.fs.isdir(path):
                self.fs.rm(path, recursive=True)
            else:
                raise ValueError(f'path is neither a direction nor a file: "{path}"')
        log.info(f'{dry_run_log_prefix}Deleted {len(path_to_delete)} files with path_props "{constraints}"')



""" This module contains the filoc default backend implementations """
from .backend_csv import CsvBackend
from .backend_json import JsonBackend
from .backend_pickle import PickleBackend
from .backend_yaml import YamlBackend

__all__     = [
    'CsvBackend',
    'JsonBackend',
    'PickleBackend',
    'YamlBackend',
]



"""
Internal utilities
"""
import json
import logging
from typing import Dict, Any, Iterable, List, Set, Union, Mapping, Tuple

from filoc.contract import PropsList

log = logging.getLogger('filoc')

_MISSING_KEY = object()


# noinspection PyMissingOrEmptyDocstring
def filter_and_coerce_loaded_file_content(path, file_content, path_props, constraints, is_singleton):
    # validate file_content and coerce to list
    if is_singleton:
        if not isinstance(file_content, dict):
            raise ValueError(f"Invalid type for loaded file '{path}'. Expected dict, got {type(file_content).__name__}")
        unfiltered_result = [file_content]
    else:
        if not isinstance(file_content, list):
            raise ValueError(f"Invalid type for loaded file '{path}'. Expected list, got {type(file_content).__name__}")
        unfiltered_result = file_content

    # filter
    result = []
    for row in unfiltered_result:
        keep = True
        for constraint_key, constraint_value in constraints.items():
            if constraint_key in path_props:
                continue  # constraint is already fulfilled by the values of the path placeholders
            if constraint_key not in row:
                continue  # constraint does not apply to this row
            if row[constraint_key] != constraint_value:
                keep = False
                break
        if keep:
            result.append(row)
    return result


# noinspection PyMissingOrEmptyDocstring
def coerce_file_content_to_write(path, props_list : PropsList, is_singleton : bool) -> Union[Mapping[str, Any], list]:
    # validate file_content and coerce to list
    if is_singleton:
        if len(props_list) == 1:
            return props_list[0]
        elif len(props_list) == 0:
            raise ValueError('props_list is empty')
        else:
            first_item = props_list[0]
            for idx, item in enumerate(props_list[1:]):
                if first_item != item:
                    raise ValueError(f'Trying to save {len(props_list)} props to singleton file {path} with different values:\nItem #0: {json.dumps(first_item)}\nItem #{idx + 1}: {json.dumps(item)}')
    else:
        return props_list


# ---------------------
# Pivot Helpers
# ---------------------

Table = List[Dict[str, Any]]
Row   = Dict[str, Any]


# noinspection PyMissingOrEmptyDocstring
def merge_tables(
    table_by_name   : Mapping[str, Table], 
    join_keys       : Iterable[str], 
    separator       : str, 
    join_level_name : str
):
    prefixed_join_keys = set([f'{join_level_name}{separator}{k}' for k in join_keys])
        
    resulting_table = None
    for table_name, table in table_by_name.items():
        table = _prefix_table(table, join_keys, table_name, separator, join_level_name)
        resulting_table = table if resulting_table is None else _join(resulting_table, table, prefixed_join_keys)
    return resulting_table


def _prefix_table(
    table          : Table, 
    join_key_names : Iterable[str],
    table_name     : str,
    separator      : str,
    join_level_name: str,
):
    result = []
    for item in table:
        result.append({  f'{join_level_name}{separator}{k}' if k in join_key_names else f'{table_name}{separator}{k}' : v  for k, v in item.items()  })
    return result


def _join(
    table1    : Table, 
    table2    : Table, 
    join_keys : Set[str]
):
    # build index on the shortest list, to speed up
    if len(table1) < len(table2):
        table1, table2 = table2, table1

    table2_index_cache = {}
    result = []
    for row1 in table1:
        rows2 = _get_rows2(row1, table2, join_keys, table2_index_cache)
        if len(rows2) == 0:
            result.append(row1.copy())  # outer join
        else:
            for row2 in rows2:
                r = row1.copy()
                r.update(row2)
                result.append(r)
    return result


def _get_rows2(
    row1      : Row, 
    table2    : Table, 
    join_keys : Set[str], 
    table2_index_cache : Dict[Tuple[str], dict]
):
    defined_keys1       = tuple(sorted(row1.keys() & join_keys))
    defined_key_values1 = [row1.get(key, _MISSING_KEY) for key in defined_keys1]

    if defined_keys1 not in table2_index_cache:
        table2_index_cache[defined_keys1] = _build_index(defined_keys1, table2)
    table2_index = table2_index_cache[defined_keys1]
    return _get_index_matches_recursive(table2_index, defined_key_values1)


def _build_index(keys : Iterable[str], table : Table) -> dict:
    index = {}
    for row in table:
        key_values = [row.get(key, _MISSING_KEY) for key in keys]
        _put_to_index(index, row, key_values)
    return index


def _put_to_index(index : dict, v : Any, key_values : List[str]):
    d_curr = index
    for k in key_values[:-1]:
        if k not in d_curr:
            d_curr[k] = {}
        d_curr = d_curr[k]

    # last is a list
    k = key_values[-1]
    if k not in d_curr:
        d_curr[k] = []
    d_curr[k].append(v)


def _get_index_matches_recursive(index, key_values):
    key_value = key_values[0]

    r1, r2 = None, None
    if len(key_values) == 1:
        if key_value    in index: r1 = index[key_value]
        if _MISSING_KEY in index: r2 = index[_MISSING_KEY]
    else:
        next_keyvalues = key_values[1:]
        if key_value    in index: r1 = _get_index_matches_recursive(index[key_value]   , next_keyvalues)
        if _MISSING_KEY in index: r1 = _get_index_matches_recursive(index[_MISSING_KEY], next_keyvalues)
    return _combine_list(r1, r2)
            

def _combine_list(l1, l2):
    if l1:
        if l2: return l1 + l2
        else : return l1
    else:
        if l2: return l2
        else : return []



""" Filoc Core Implementation """
import json
import logging
import os, sys
import pickle
import random
import socket
import threading
import time
from abc import ABC
from contextlib import contextmanager
from datetime import datetime
from io import UnsupportedOperation
from typing import Dict, Any, NamedTuple, Optional, Tuple, Set

from frozendict import frozendict
from fsspec.spec import AbstractFileSystem

from filoc.contract import TContent, TContents, Constraints, Props, PropsList, Filoc, \
    FrontendContract, BackendContract, ReadOnlyProps, Constraint, ReadOnlyPropsList
from .filoc_io import FilocIO, mix_dicts_and_coerce
from .utils import merge_tables

log = logging.getLogger('filoc')


class LockException(Exception):
    """ Exception raised while trying to acquire a lock with ``Filoc.lock()`` after the count of defined attempts has been reached"""
    def __init__(self, *args):
        Exception.__init__(self, *args)


class _RunningCache(NamedTuple):
    key : Props  # key-values expected by cache_locpath
    cache_by_file_path_props : Dict[ReadOnlyProps, Dict[str, Any]]  # key-values expected by (this) locpath props


# TODO: Profile and apply intern(key) to reduce footprint of intermediate model dictionaries
# TODO Feature: optimistic locking for long editing (use timestamp_col)
# TODO Feature: Backup on write if file already exists (think of possibility to use locpath with $version and $timestamp placeholder)

# ---------
# FilocSingle
# ---------
class FilocSingle(FilocIO, Filoc[TContent, TContents], ABC):
    """ Filoc implementation for a single locpath """

    # noinspection PyDefaultArgument
    def __init__(
            self,
            locpath       : str                                  ,
            writable      : bool                                 ,
            transaction   : bool                                 ,
            frontend      : FrontendContract[TContent, TContents],
            backend       : BackendContract                      ,
            cache_locpath : str                                  ,
            cache_fs      : Optional[AbstractFileSystem]         ,
            timestamp_col : str                                  ,
            fs            : Optional[AbstractFileSystem]         ,
    ):
        """
        if cache_locpath is relative, then it will be relative to result_locpath
        """
        FilocIO.__init__(self, locpath, writable, fs)
        self._transaction   = transaction
        self._frontend      = frontend
        self._backend       = backend

        # cache loc
        self._cache_loc = None
        if cache_locpath is not None:
            self._cache_loc = FilocIO(cache_locpath, writable=True, fs=cache_fs)
        self._timestamp_col = timestamp_col

    @contextmanager
    def lock(self, attempt_count: int = 60, attempt_secs: float = 1.0):
        """ See ``Filoc`` contract """
        # remark: fsspec backends do not all support the 'x' exclusive mode, so we cannot use exclusive write mode to
        # synchronize the writing into a single lock file. So each call to `lock()` tries to write its own lock file
        # and check afterward, if he won the run, by checking the modified timestamp of all lock files. It assumes, that the
        # file system sets timestamps in the same order as it processes the files (TODO: Verify assumption on distributed file systems)

        lock_id, lock_file = self._get_my_lock_id_and_lock_file()
        for attempt in range(attempt_count):
            owning_lock_date_and_file = self._get_owning_lock_date_and_file()

            if owning_lock_date_and_file:
                if owning_lock_date_and_file[1].endswith(lock_id): 
                    yield lock_id
                    return
                else:
                    time.sleep(random.uniform(0.5 * attempt_secs, 1.5 * attempt_secs))
                    continue

            # else we try to acquire the lock
            self.fs.makedirs(self._root_folder, exist_ok=True)
            with self.fs.open(lock_file, 'w') as f:
                json.dump({
                    'host' : socket.gethostname(),
                    'pid' : os.getpid(),
                    'thread' : threading.get_ident(),
                }, f)

            try:
                owning_lock_date_and_file = self._get_owning_lock_date_and_file()
                if owning_lock_date_and_file and owning_lock_date_and_file[1].endswith(lock_id):
                    yield lock_id
                    return
            finally:
                # else either failed to acquire lock (concurrent won) or some error. We clean up and retry (loop)
                try:
                    self.fs.delete(lock_file)
                except FileNotFoundError:
                    log.warning(f"Lock file {lock_file} has been concurrently deleted (by self.lock_force_release()?). No need to remove it")        
        raise LockException(f"Failed to acquire the file lock after {attempt_count} attempts")

    def lock_info(self) -> Optional[Dict[str, Any]]:
        """ See ``Filoc`` contract """
        owning_lock_date_and_file = self._get_owning_lock_date_and_file()
        if owning_lock_date_and_file is None:
            return None

        try:
            with self.fs.open(owning_lock_date_and_file[1], 'r') as f:
                info = json.load(f)
            info['date'] = owning_lock_date_and_file[0]
            return info
        except FileNotFoundError:
            return None

    def lock_force_release(self):
        """ See ``Filoc`` contract """
        owning_lock_date_and_file = self._get_owning_lock_date_and_file()
        if owning_lock_date_and_file is None:
            log.info(f'No lock found')
            return

        lock_file = owning_lock_date_and_file[1]
        try:
            self.fs.delete(lock_file)
            log.warning(f'Forced releasing of lock file "{lock_file}"')
        except FileNotFoundError:
            return

    def _get_owning_lock_date_and_file(self) -> Optional[Tuple[datetime, str]]:
        lock_files = self.fs.glob(f'{self._root_folder}/.lock_*')
        if len(lock_files) == 0:
            return None
        
        oldest_date = None
        oldest_file = None
        for lock_file in lock_files:
            try: 
                date = self.fs.modified(lock_file)
            except FileNotFoundError: 
                continue
            if oldest_date is None or date < oldest_date:
                oldest_date = date
                oldest_file = lock_file
        if oldest_file is None:
            return None
        return oldest_date, oldest_file

    def _get_my_lock_id_and_lock_file(self):
        # build a lock id. Scope of lock is (process x thread x root_folder)
        host      = socket.gethostname()
        pid       = os.getpid()
        thread_id = threading.get_ident()
        lock_id   = f'{host}_{pid}_{thread_id}'
        lock_file = f'{self._root_folder}/.lock_{lock_id}'
        return lock_id, lock_file

    def invalidate_cache(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint):
        """ See ``Filoc`` contract """
        if self._cache_loc is None:
            return
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        self._cache_loc.delete(constraints)

    def read_content(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> TContent:
        """ See ``Filoc`` contract """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        props_list = self._read_props_list(constraints)
        return self._frontend.read_content(props_list)

    def read_contents(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> TContents:
        """ See ``Filoc`` contract """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        props_list  = self._read_props_list(constraints)
        return self._frontend.read_contents(props_list)

    def _read_props_list(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> PropsList:
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        result = []

        running_cache = None  # type:Optional[_RunningCache]

        paths_and_file_path_props  = self.list_paths_and_props(constraints)
        log.info(f'Found {len(paths_and_file_path_props)} files to read in locpath {self._locpath} fulfilling props {constraints}')

        for (path, file_path_props) in paths_and_file_path_props:
            path_props_hashable = frozendict(file_path_props.items())

            try:
                f_timestamp = self.fs.modified(path)
            except NotImplementedError:
                # fsspec implementation, that do not implement modified, are assumed to be read-only (example: github)
                f_timestamp = None
            except FileNotFoundError:
                f_timestamp = None
                
            # renew cache, on cache file change
            if self._cache_loc:
                path_cache_path       = self._cache_loc.render_path(file_path_props)
                cache_file_path_props = self._cache_loc.parse_path_properties(path_cache_path)

                if running_cache is not None and running_cache.key == cache_file_path_props:
                    pass  # cache file is always the correct one : do nothing, keep this cache file opened
                else:
                    if running_cache is not None and running_cache.key != cache_file_path_props:
                        # flush previous cache, if exists
                        if running_cache.key:
                            with self._cache_loc.open(running_cache.key, 'wb') as f:
                                pickle.dump(running_cache.cache_by_file_path_props, f, )

                    # now prepare new cache file
                    if self._cache_loc.exists(cache_file_path_props):
                        with self._cache_loc.open(cache_file_path_props, 'rb') as f:
                            running_cache = _RunningCache(cache_file_path_props, pickle.load(f))
                    else:
                        running_cache = _RunningCache(cache_file_path_props, dict())

            # check whether cache entry is still valid
            if self._cache_loc:
                if path_props_hashable in running_cache.cache_by_file_path_props:
                    path_cached_entry = running_cache.cache_by_file_path_props[path_props_hashable]
                    if path_cached_entry['timestamp'] == f_timestamp:
                        log.info(f'Path data cached: "{path}"')
                        result.extend(path_cached_entry['props_list'].copy())  # copy from cache
                        continue
                    else:
                        log.info(f'Cache out of date for path "{path}"')

            # cache is not valid: read path directly!

            # props from reader
            props_list = self._read_path(path, file_path_props, constraints)    # type: PropsList

            # augment read props with additional external data
            for props in props_list:
                # -> file timestamp
                if self._timestamp_col:
                    props[self._timestamp_col] = f_timestamp
                # -> path_props
                props.update(file_path_props)

            # add to result
            result.extend(props_list)
            # add to cache
            if self._cache_loc:
                running_cache.cache_by_file_path_props[path_props_hashable] = { 'timestamp': f_timestamp, 'props_list' : props_list.copy()}

        # flush last used cache
        if running_cache:
            with self._cache_loc.open(running_cache.key, 'wb') as f:
                pickle.dump(running_cache.cache_by_file_path_props, f)

        return result

    def write_content(self, content : TContent, dry_run=False):
        """ See ``Filoc`` contract """
        if not self._writable:
            raise UnsupportedOperation('this filoc is not writable. Set writable flag to True to enable writing')
        
        props_list = self._frontend.write_content(content)

        def save():
            self._write_props_list(props_list, dry_run=dry_run)

        if self._transaction:
            with self.fs.transaction:
                save()
        else:
            save()

    def write_contents(self, contents : TContents, dry_run=False):
        """ See ``Filoc`` contract """
        if not self._writable:
            raise UnsupportedOperation('this filoc is not writable. Set writable flag to True to enable writing')

        props_list = self._frontend.write_contents(contents)

        def save():
            self._write_props_list(props_list, dry_run=dry_run)

        if self._transaction:
            with self.fs.transaction:
                save()
        else:
            save()

    def _write_props_list(self, props_list : ReadOnlyPropsList, dry_run=False):
        if not self._writable:
            raise UnsupportedOperation('this filoc is not writable. Set writable flag to True to enable writing')

        recorded_row_id_by_path_props    = {}
        recorded_row_other_props_by_path_props = {}
        for row_id, row_props in enumerate(props_list):
            path_props, row_other_props, row_timestamp = self._split_keyvalues(row_props)
            row_path_props_hashable = frozendict(path_props)

            if row_path_props_hashable not in recorded_row_other_props_by_path_props:
                recorded_row_other_props_by_path_props[row_path_props_hashable] = []
                recorded_row_id_by_path_props[row_path_props_hashable] = []

            # record change to perform and notice row_id (last is for error message)
            recorded_row_id_by_path_props[row_path_props_hashable].append(row_id)
            recorded_row_other_props_by_path_props[row_path_props_hashable].append(row_other_props)

        dry_run_log_prefix = '(dry_run) ' if dry_run else ''
        for path_props, other_props_list in recorded_row_other_props_by_path_props.items():
            self.invalidate_cache(path_props)
            path = self.render_path(path_props)

            log.info(f'{dry_run_log_prefix}Saving to {path}')
            if not dry_run:
                self._backend.write(self.fs, path, other_props_list)
            log.info(f'{dry_run_log_prefix}Saved {path}')

    def _split_keyvalues(self, keyvalues : ReadOnlyProps) -> Tuple[Props, Props, datetime]:
        path_props  = {}
        timestamp   = None
        other_props = dict()
        for (k, v) in keyvalues.items():
            if k in self._path_props:
                path_props[k] = v
            elif k == self._timestamp_col:
                timestamp = keyvalues[k]
            else:
                other_props[k] = v
        return path_props, other_props, timestamp

    def _read_path(self, path : str, path_props : Props, constraints : Constraints):
        log.info(f'Reading content for {path}')
        content = self._backend.read(self.fs, path, path_props, constraints)
        log.info(f'Read content for {path}')
        return content


# ------------------
# FilocCompositeBase
# ------------------
class FilocComposite(Filoc[TContent, TContents], ABC):
    """ Filoc implementation for composite filocs """

    # TODO: Implement remaining Filoc methods (lock() and co.)

    def __init__(
            self,
            filoc_by_name           : Dict[str, FilocSingle],
            frontend                : FrontendContract,
            transaction             : bool,
            join_level_name         : str,
            join_separator          : str,
    ):
        # Validate
        assert isinstance(filoc_by_name, dict)
        for filoc_name, filoc in filoc_by_name.items():
            if not isinstance(filoc, FilocSingle):
                raise ValueError(f'filoc {filoc_name} is not a FilocSingle instance. FilocComposite only support FilocSingle sub-filocs')

        self.frontend        = frontend
        self.transaction     = transaction
        self.filoc_by_name   = filoc_by_name
        self.join_level_name = join_level_name
        self.join_separator  = join_separator

        self.join_keys_by_filoc_name = {}  # type: Dict[str, Set[str]]
        for filoc_name, filoc in filoc_by_name.items():
            self.join_keys_by_filoc_name[filoc_name] = filoc._path_props

    def invalidate_cache(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint):
        """ see ``Filoc`` contract """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        for filoc in self.filoc_by_name.values():
            filoc.invalidate_cache(constraints)

    def read_content(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> TContent:
        """ see ``Filoc`` contract """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        props_list = self._read_props_list(constraints)
        return self.frontend.read_content(props_list)

    def read_contents(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> TContents:
        """ see ``Filoc`` contract """
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        props_list = self._read_props_list(constraints)
        return self.frontend.read_contents(props_list)

    def _read_props_list(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> PropsList:
        constraints = mix_dicts_and_coerce(constraints, constraints_kwargs)
        # collect
        props_list_by_filoc_name = {}
        for filoc_name, filoc in self.filoc_by_name.items():
            # noinspection PyProtectedMember
            props_list_by_filoc_name[filoc_name] = filoc._read_props_list(constraints)

        # join
        join_key_names = set()
        for filoc_join_key_names in self.join_keys_by_filoc_name.values():
            join_key_names = join_key_names | filoc_join_key_names
        return merge_tables(props_list_by_filoc_name, list(join_key_names), self.join_separator, self.join_level_name)

    def write_content(self, content : TContent, dry_run=False):
        """ see ``Filoc`` contract """
        props_list = self.frontend.write_content(content)
        self._write_props_list(props_list, dry_run=dry_run)

    def write_contents(self, contents : TContents, dry_run=False):
        """ see ``Filoc`` contract """
        props_list = self.frontend.write_contents(contents)
        self._write_props_list(props_list, dry_run=dry_run)

    def _write_props_list(self, props_list : ReadOnlyPropsList, dry_run=False):
        # prepare empty props_list by filoc_name
        props_list_by_filoc_name = {}
        props_list_by_filoc_name[self.join_level_name] = [ dict() for _ in range(len(props_list)) ]
        for filoc_name, filoc in self.filoc_by_name.items():
            if filoc._writable:
                props_list_by_filoc_name[filoc_name] = [ dict() for _ in range(len(props_list)) ]
            else:
                log.info(f'write operation skipped for "{filoc_name}" readonly Filoc')

        # then fill
        split_cache = {}
        for row_id, props in enumerate(props_list):
            for (k, v) in props.items():

                split_values = split_cache.get(k, None)
                if split_values is None:
                    split_values = k.split(self.join_separator, 1)
                    split_cache[k] = split_values

                filoc_name, prop_name = split_values

                if filoc_name not in props_list_by_filoc_name:
                    continue

                props_list_by_filoc_name[filoc_name][row_id][prop_name] = v

        # pop out join indexes and merge then row by row to filoc props
        indexes = props_list_by_filoc_name.pop(self.join_level_name)
        for filoc_name, filoc_props_list in props_list_by_filoc_name.items():
            join_keys = self.join_keys_by_filoc_name[filoc_name]
            for index, props in zip(indexes, filoc_props_list):
                # todo: iterate through join_keys instead of index and raise explicit exception if join_key missing?
                relevant_index = { k: v for (k, v) in index.items() if k in join_keys}
                props.update(relevant_index)

        # delegate writing to
        def save():
            for filoc_name, filoc_props_list in props_list_by_filoc_name.items():
                filoc = self.filoc_by_name[filoc_name]
                # noinspection PyProtectedMember
                filoc._write_props_list(filoc_props_list, dry_run=dry_run)

        if self.transaction:

            # begin compound transaction
            for filoc_name in props_list_by_filoc_name:
                filoc = self.filoc_by_name[filoc_name]
                filoc.fs.transaction.__enter__()

            try:
                save()

                # commit compound transaction
                for filoc_name in props_list_by_filoc_name:
                    filoc = self.filoc_by_name[filoc_name]
                    filoc.fs.transaction.__exit__(None, None, None)

            except:
                exc_info = sys.exc_info()
                # rollback compound transaction
                for filoc_name in props_list_by_filoc_name:
                    filoc = self.filoc_by_name[filoc_name]
                    filoc.fs.transaction.__exit__(*exc_info)
        else:
            # no transaction
            save()

    @contextmanager
    def lock(self, attempt_count: int = 60, attempt_secs: float = 1.0):
        """ See ``Filoc`` contract """
        raise NotImplementedError("TODO: Implement")

    def lock_info(self) -> Optional[Dict[str, Any]]:
        """ See ``Filoc`` contract """
        raise NotImplementedError("TODO: Implement")

    def lock_force_release(self):
        """ See ``Filoc`` contract """
        raise NotImplementedError("TODO: Implement")



import json
import shutil
import tempfile
import unittest

from pandas import DataFrame

from filoc import filoc


# noinspection PyMissingOrEmptyDocstring
class TestBackendCsv(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')
        self.file = self.test_dir + r'/myfile.csv'

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read_contents(self):
        data = {'a': 123, 'b': 'coucou'}
        with open(self.file, mode='w') as f:
            f.write('a,b\n123,coucou')

        loc = filoc(self.file, backend='csv', singleton=True)
        df = loc.read_contents()
        col_names = sorted(df.columns)
        self.assertEqual(col_names, ['a', 'b'])
        self.assertEqual(df.loc[0, 'a'], '123')
        self.assertEqual(df.loc[0, 'b'], 'coucou')

    def test_write_contents(self):
        data = {'a': 123, 'b': 'coucou'}
        df = DataFrame([data])

        loc = filoc(self.file, backend='csv', singleton=True, writable=True)
        df = loc.write_contents(df)

        with open(self.file) as f:
            content = f.read()
        rows = content.split('\n')
        data = list(filter(None, [list(filter(None, r.split(','))) for r in rows]))
        self.assertEqual([['a', 'b'], ['123', 'coucou']], data)

    def test_read_file_with_bom(self):
        with open(self.file, mode='w', encoding='utf-8-sig') as f:
            f.writelines(['A,B\n1,2\n3,4'])

        loc = filoc(self.file, writable=True, backend='csv', singleton=False, encoding='utf-8-sig')
        df = loc.read_contents()
        col_name = df.columns[0]
        self.assertEqual(col_name, 'A')

    def test_read_file_no_bom(self):
        with open(self.file, mode='w', encoding='utf-8') as f:
            f.writelines(['A,B\n1,2\n3,4'])

        loc = filoc(self.file, writable=True, backend='csv', singleton=False, encoding='utf-8')
        df = loc.read_contents()
        col_name = df.columns[0]
        self.assertEqual(col_name, 'A')

if __name__ == '__main__':
    unittest.main()



import json
import shutil
import tempfile
from typing import Any, List
import unittest

# noinspection DuplicatedCode
from filoc import filoc_json


# noinspection PyMissingOrEmptyDocstring
class TestMultiloc_TwoLevels(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read(self):
        node_loc = filoc_json(f'{self.test_dir}/{{node_id:d}}/node.json', writable=True)
        leaf_loc = filoc_json(f'{self.test_dir}/{{node_id:d}}/{{leaf_id:d}}/leaf.json', writable=True)
        sut_loc = filoc_json({
            'node' : node_loc,
            'leaf' : leaf_loc,
        })

        node_loc.write_content( {'node_id': 1, 'value' : 'A'})
        node_loc.write_content( {'node_id': 2, 'value' : 'B'})
        leaf_loc.write_content( {'node_id': 1, 'leaf_id': 1, 'value' : 0.1})
        leaf_loc.write_content( {'node_id': 1, 'leaf_id': 2, 'value' : 0.2})
        leaf_loc.write_content( {'node_id': 2, 'leaf_id': 1, 'value' : 1.1})
        leaf_loc.write_content( {'node_id': 2, 'leaf_id': 2, 'value' : 1.2})

        result = sut_loc.read_contents()

        result_tree = pivot(result, "shared.node_id", "shared.leaf_id")
        result_tree_txt = json.dumps(result_tree, sort_keys=True)
        print(result_tree_txt)

        self.maxDiff = 2000
        self.assertEqual(result_tree_txt, 
            '{'
            '"1": {'
              '"1": {"leaf.value": 0.1, "node.value": "A", "shared.leaf_id": 1, "shared.node_id": 1}, '
              '"2": {"leaf.value": 0.2, "node.value": "A", "shared.leaf_id": 2, "shared.node_id": 1}'
            '}, '
            '"2": {'
              '"1": {"leaf.value": 1.1, "node.value": "B", "shared.leaf_id": 1, "shared.node_id": 2}, '
              '"2": {"leaf.value": 1.2, "node.value": "B", "shared.leaf_id": 2, "shared.node_id": 2}'
              '}'
            '}'
        )

    def test_2filocs_with_reversed_key_order(self):
        leaf1_loc = filoc_json(f'{self.test_dir}/{{leaf_id:d}}/{{node_id:d}}/leaf1.json', writable=True)
        leaf2_loc = filoc_json(f'{self.test_dir}/{{node_id:d}}/{{leaf_id:d}}/leaf2.json', writable=True)
        sut_loc = filoc_json({ 'leaf1' : leaf1_loc, 'leaf2' : leaf2_loc })

        leaf1_loc.write_content( {'node_id': 1, 'leaf_id': 1, 'value' : 0.1})
        leaf1_loc.write_content( {'node_id': 1, 'leaf_id': 2, 'value' : 0.2})
        leaf1_loc.write_content( {'node_id': 2, 'leaf_id': 1, 'value' : 1.1})
        leaf1_loc.write_content( {'node_id': 2, 'leaf_id': 2, 'value' : 1.2})
        leaf2_loc.write_content( {'node_id': 1, 'leaf_id': 1, 'value' : 1.0})
        leaf2_loc.write_content( {'node_id': 1, 'leaf_id': 2, 'value' : 2.0})
        leaf2_loc.write_content( {'node_id': 2, 'leaf_id': 1, 'value' : 11.0})
        leaf2_loc.write_content( {'node_id': 2, 'leaf_id': 2, 'value' : 12.0})

        result = sut_loc.read_contents()

        result_tree = pivot(result, "shared.node_id", "shared.leaf_id")
        result_tree_txt = json.dumps(result_tree, sort_keys=True)
        print(result_tree_txt)

        self.maxDiff = 2000
        self.assertEqual(result_tree_txt, 
            '{'
            '"1": {'
              '"1": {"leaf1.value": 0.1, "leaf2.value": 1.0, "shared.leaf_id": 1, "shared.node_id": 1}, '
              '"2": {"leaf1.value": 0.2, "leaf2.value": 2.0, "shared.leaf_id": 2, "shared.node_id": 1}'
            '}, '
            '"2": {'
              '"1": {"leaf1.value": 1.1, "leaf2.value": 11.0, "shared.leaf_id": 1, "shared.node_id": 2}, '
              '"2": {"leaf1.value": 1.2, "leaf2.value": 12.0, "shared.leaf_id": 2, "shared.node_id": 2}'
              '}'
            '}'
        )

    def test_2filocs_with_unrelated_placeholder_keys(self):
        leaf1_loc = filoc_json(f'{self.test_dir}/1/{{A}}/{{B}}/leaf1.json', writable=True)
        leaf2_loc = filoc_json(f'{self.test_dir}/2/{{C}}/{{D}}/leaf2.json', writable=True)
        sut_loc = filoc_json({ 'AB' : leaf1_loc, 'CD' : leaf2_loc })

        leaf1_loc.write_content( {'A': 1  , 'B':   10, 'V' : 0.1})
        leaf1_loc.write_content( {'A': 1  , 'B':   20, 'V' : 0.2})
        leaf2_loc.write_content( {'C': 100, 'D': 1000, 'V' : 1.0})
        leaf2_loc.write_content( {'C': 100, 'D': 2000, 'V' : 2.0})

        result = sut_loc.read_contents()

        result_tree = pivot(result, "shared.A", "shared.B", "shared.C", "shared.D")
        result_tree_txt = json.dumps(result_tree, sort_keys=True)
        print(result_tree_txt)

        self.maxDiff = 2000
        self.assertEqual(result_tree_txt, '{"1": {"10": {"100": {"1000": {"AB.V": 0.1, "CD.V": 1.0, "shared.A": "1", "shared.B": "10", "shared.C": "100", "shared.D": "1000"}, "2000": {"AB.V": 0.1, "CD.V": 2.0, "shared.A": "1", "shared.B": "10", "shared.C": "100", "shared.D": "2000"}}}, "20": {"100": {"1000": {"AB.V": 0.2, "CD.V": 1.0, "shared.A": "1", "shared.B": "20", "shared.C": "100", "shared.D": "1000"}, "2000": {"AB.V": 0.2, "CD.V": 2.0, "shared.A": "1", "shared.B": "20", "shared.C": "100", "shared.D": "2000"}}}}}')

    def test_2filocs_second_is_empty(self):
        leaf1_loc = filoc_json(f'{self.test_dir}/1/{{A}}/{{B}}/leaf1.json', writable=True)
        leaf2_loc = filoc_json(f'{self.test_dir}/2/{{C}}/{{D}}/leaf2.json', writable=True)
        sut_loc = filoc_json({ 'AB' : leaf1_loc, 'CD' : leaf2_loc })

        leaf2_loc.write_content( {'C': 100, 'D': 1000, 'V' : 1.0})
        leaf2_loc.write_content( {'C': 100, 'D': 2000, 'V' : 2.0})

        result = sut_loc.read_contents()

        result_tree = pivot(result, "shared.C", "shared.D")
        result_tree_txt = json.dumps(result_tree, sort_keys=True)
        print(result_tree_txt)

        self.maxDiff = 2000
        self.assertEqual(result_tree_txt, '{"100": {"1000": {"CD.V": 1.0, "shared.C": "100", "shared.D": "1000"}, "2000": {"CD.V": 2.0, "shared.C": "100", "shared.D": "2000"}}}')

    def test_2filocs_first_is_empty(self):
        leaf1_loc = filoc_json(f'{self.test_dir}/1/{{A}}/{{B}}/leaf1.json', writable=True)
        leaf2_loc = filoc_json(f'{self.test_dir}/2/{{C}}/{{D}}/leaf2.json', writable=True)
        sut_loc = filoc_json({ 'AB' : leaf1_loc, 'CD' : leaf2_loc })

        leaf1_loc.write_content( {'A': 1  , 'B':   10, 'V' : 0.1})
        leaf1_loc.write_content( {'A': 1  , 'B':   20, 'V' : 0.2})

        result = sut_loc.read_contents()

        result_tree = pivot(result, "shared.A", "shared.B")
        result_tree_txt = json.dumps(result_tree, sort_keys=True)
        print(result_tree_txt)

        self.maxDiff = 2000
        self.assertEqual(result_tree_txt, '{"1": {"10": {"AB.V": 0.1, "shared.A": "1", "shared.B": "10"}, "20": {"AB.V": 0.2, "shared.A": "1", "shared.B": "20"}}}'
        )

    def test_2filocs_filter_on_first_locpath(self):
        leaf1_loc = filoc_json(f'{self.test_dir}/{{node_id:d}}/{{leaf_id:d}}/leaf1.json', writable=True)
        leaf2_loc = filoc_json(f'{self.test_dir}/{{node_id:d}}/{{leaf_id:d}}/leaf2.json', writable=True)
        sut_loc = filoc_json({ 'leaf1' : leaf1_loc, 'leaf2' : leaf2_loc })

        leaf1_loc.write_content( {'node_id': 1, 'leaf_id': 1, 'value' : 0.1})
        leaf1_loc.write_content( {'node_id': 1, 'leaf_id': 2, 'value' : 0.2})
        leaf1_loc.write_content( {'node_id': 2, 'leaf_id': 1, 'value' : 1.1})
        leaf1_loc.write_content( {'node_id': 2, 'leaf_id': 2, 'value' : 1.2})
        leaf2_loc.write_content( {'node_id': 1, 'leaf_id': 1, 'value' : 1.0})
        leaf2_loc.write_content( {'node_id': 1, 'leaf_id': 2, 'value' : 2.0})
        leaf2_loc.write_content( {'node_id': 2, 'leaf_id': 1, 'value' : 11.0})
        leaf2_loc.write_content( {'node_id': 2, 'leaf_id': 2, 'value' : 12.0})

        result = sut_loc.read_contents()

        result_tree = pivot(result, "shared.node_id", "shared.leaf_id")
        result_tree_txt = json.dumps(result_tree, sort_keys=True)
        print(result_tree_txt)

        self.maxDiff = 2000
        self.assertEqual(result_tree_txt, 
            '{'
            '"1": {'
              '"1": {"leaf1.value": 0.1, "leaf2.value": 1.0, "shared.leaf_id": 1, "shared.node_id": 1}, '
              '"2": {"leaf1.value": 0.2, "leaf2.value": 2.0, "shared.leaf_id": 2, "shared.node_id": 1}'
            '}, '
            '"2": {'
              '"1": {"leaf1.value": 1.1, "leaf2.value": 11.0, "shared.leaf_id": 1, "shared.node_id": 2}, '
              '"2": {"leaf1.value": 1.2, "leaf2.value": 12.0, "shared.leaf_id": 2, "shared.node_id": 2}'
              '}'
            '}'
        )

def pivot(l : List[dict], *keys):
    result = {}
    for item in l:
        path_values = [item[k] for k in keys]
        put_value(result, item, *path_values)
    return result


def put_value(d : dict, v : Any, *path_values : str):
    d_curr = d
    for k in path_values[:-1]:
        if k not in d_curr:
            d_curr[k] = {}
        d_curr = d_curr[k]
    d_curr[path_values[-1]] = v


if __name__ == '__main__':
    unittest.main()



""" github protocol test """
import os
import shutil
import tempfile
import unittest
from pathlib import Path

from filoc import filoc
from filoc import utils

# noinspection PyMissingOrEmptyDocstring
def touch_file(file_path):
    os.makedirs(os.path.dirname(file_path))
    Path(file_path).touch()


# noinspection PyMissingOrEmptyDocstring,PyPep8Naming
class TestUtils(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test__combine_list_None_None(self):
        r = utils._combine_list(None, None)
        self.assertEqual([], r)

    def test__combine_list_None_ok(self):
        r = utils._combine_list(None, [1, 2])
        self.assertEqual([1, 2], r)

    def test__combine_list_ok_None(self):
        r = utils._combine_list([1, 2], None)
        self.assertEqual([1, 2], r)

    def test__combine_list_ok_ok(self):
        r = utils._combine_list([1, 2], [3, 4])
        self.assertEqual([1, 2, 3, 4], r)



import json
import yaml
import shutil
import tempfile
import unittest
from pandas import DataFrame
from filoc import filoc

# noinspection PyMissingOrEmptyDocstring
class TestBackendYaml(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')
        self.file = self.test_dir + r'/myfile.yaml'

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read_contents(self):
        data = {'a': 123, 'b': [1, 2, 3]}
        with open(self.file, mode='w') as f:
            yaml.dump(data, f)

        loc = filoc(self.file, backend='yaml', singleton=True)
        df = loc.read_contents()
        col_names = sorted(df.columns)
        self.assertEqual(col_names, ['a', 'b'])
        self.assertEqual(df.loc[0, 'a'], 123)

    def test_write_contents(self):
        data = {'a': 123, 'b': [1, 2, 3]}
        df = DataFrame([data])

        loc = filoc(self.file, backend='yaml', singleton=True, writable=True)
        df = loc.write_contents(df)

        with open(self.file) as f:
            content = yaml.safe_load(f)

        self.assertEqual(json.dumps(content, sort_keys=True),
                         json.dumps(data, sort_keys=True))


if __name__ == '__main__':
    unittest.main()



""" Test for json frontend """
import json
import os
import shutil
import tempfile
import time
import unittest
from pathlib import Path

from filoc import filoc_json, FilocIO


# noinspection PyMissingOrEmptyDocstring
from filoc.contract import SingletonExpectedError


def touch_file(file_path):
    os.makedirs(os.path.dirname(file_path))
    Path(file_path).touch()


# noinspection DuplicatedCode
# noinspection PyMissingOrEmptyDocstring
class TestFilocSingle(unittest.TestCase):
    """
    TODO: 
    - Test cache behavior on delete of files (should currently fail --> TODO DEV Feature)    
    """

    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')
        self.path_fmt = self.test_dir + r'/simid={simid:d}/epid={epid:d}/hyperparameters.json'

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read_write_read(self):
        wloc = FilocIO(self.path_fmt, writable=True)
        with wloc.open({"simid": 1, "epid": 10}, "w") as f: json.dump({'a': 100}, f)
        with wloc.open({"simid": 1, "epid": 20}, "w") as f: json.dump({'a': 200}, f)
        with wloc.open({"simid": 2, "epid": 10}, "w") as f: json.dump({'a': 300}, f)
        with wloc.open({"simid": 2, "epid": 20}, "w") as f: json.dump({'a': 400}, f)

        loc = filoc_json(self.path_fmt)
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 300, "epid": 10, "simid": 2}]', json.dumps(p, sort_keys=True))

        # change file
        time.sleep(0.1)  # ensures different timestamp
        with wloc.open({"simid": 2, "epid": 10}, "w") as f:
            json.dump({'a': 333}, f)
        
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 333, "epid": 10, "simid": 2}]', json.dumps(p, sort_keys=True))

    def test_read_all(self):
        wloc = FilocIO(self.path_fmt, writable=True)
        with wloc.open({"simid": 1, "epid": 10}, "w") as f: json.dump({'a': 100}, f)
        with wloc.open({"simid": 1, "epid": 20}, "w") as f: json.dump({'a': 200}, f)
        with wloc.open({"simid": 2, "epid": 10}, "w") as f: json.dump({'a': 300}, f)
        with wloc.open({"simid": 2, "epid": 20}, "w") as f: json.dump({'a': 400}, f)

        loc = filoc_json(self.path_fmt)
        p = loc.read_contents()
        self.assertEqual(len(p), 4)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 200, "epid": 20, "simid": 1}, {"a": 300, "epid": 10, "simid": 2}, {"a": 400, "epid": 20, "simid": 2}]', json.dumps(p, sort_keys=True))

    def test_with_constraints_on_path_placeholders(self):
        wloc = FilocIO(self.path_fmt, writable=True)
        with wloc.open({"simid": 1, "epid": 10}, "w") as f: json.dump({'a': 100}, f)
        with wloc.open({"simid": 1, "epid": 20}, "w") as f: json.dump({'a': 200}, f)
        with wloc.open({"simid": 2, "epid": 10}, "w") as f: json.dump({'a': 300}, f)
        with wloc.open({"simid": 2, "epid": 20}, "w") as f: json.dump({'a': 400}, f)

        loc = filoc_json(self.path_fmt)
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 300, "epid": 10, "simid": 2}]', json.dumps(p, sort_keys=True))

    def test_with_constraints_on_content_attributes(self):
        wloc = FilocIO(self.path_fmt, writable=True)
        with wloc.open({"simid": 1, "epid": 10}, "w") as f: json.dump({'a': 100}, f)
        with wloc.open({"simid": 1, "epid": 20}, "w") as f: json.dump({'a': 200}, f)
        with wloc.open({"simid": 2, "epid": 10}, "w") as f: json.dump({'a': 300}, f)
        with wloc.open({"simid": 2, "epid": 20}, "w") as f: json.dump({'a': 400}, f)

        loc = filoc_json(self.path_fmt)
        p = loc.read_contents({'a': 300})
        self.assertEqual(len(p), 1)
        self.assertEqual('[{"a": 300, "epid": 10, "simid": 2}]', json.dumps(p, sort_keys=True))

    def test_read_contents_with_cache(self):
        print("write files")
        wloc = FilocIO(self.path_fmt, writable=True)
        with wloc.open({"simid": 1, "epid": 10}, "w") as f:
            json.dump({'a': 100}, f)
        with wloc.open({"simid": 1, "epid": 20}, "w") as f:
            json.dump({'a': 200}, f)
        with wloc.open({"simid": 2, "epid": 10}, "w") as f:
            json.dump({'a': 300}, f)
        with wloc.open({"simid": 2, "epid": 20}, "w") as f:
            json.dump({'a': 400}, f)

        loc = filoc_json(self.path_fmt, cache_locpath='.cache')
        print("read_contents 1")
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 300, "epid": 10, "simid": 2}]',
                         json.dumps(p, sort_keys=True))

        time.sleep(0.1)  # small delay between the two writings to ensure that the file gets an new timestamp

        print("change one file")
        # change to file triggers cache refresh
        with wloc.open({"simid": 2, "epid": 10}, "w") as f: 
            json.dump({'a': 333}, f)
            # f.flush()
            # os.fsync(f.fileno())

        print("re read_contents 2")
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 333, "epid": 10, "simid": 2}]', json.dumps(p, sort_keys=True))

        # act_assert()

        # Trick to test: signature change does not take effect, because of cache
        loc = filoc_json(self.path_fmt, cache_locpath='.cache')
        print("re read_contents 3")
        p = loc.read_contents({'epid': 10})
        self.assertEqual(len(p), 2)
        self.assertEqual('[{"a": 100, "epid": 10, "simid": 1}, {"a": 333, "epid": 10, "simid": 2}]',
                         json.dumps(p, sort_keys=True))

    def test_write_contents(self):
        wloc = filoc_json(self.path_fmt, writable=True)
        wloc._write_props_list([
            {"simid": 1, "epid": 10, 'a': 100},
            {"simid": 1, "epid": 20, 'a': 200},
            {"simid": 2, "epid": 10, 'a': 300},
            {"simid": 2, "epid": 20, 'a': 400},
        ])

        wloc = FilocIO(self.path_fmt)
        with wloc.open({"simid": 1, "epid": 10}) as f:
            c1 = json.load(f)
        with wloc.open({"simid": 1, "epid": 20}) as f:
            c2 = json.load(f)
        with wloc.open({"simid": 2, "epid": 10}) as f:
            c3 = json.load(f)
        with wloc.open({"simid": 2, "epid": 20}) as f:
            c4 = json.load(f)

        self.assertEqual('{"a": 100}', json.dumps(c1, sort_keys=True))
        self.assertEqual('{"a": 200}', json.dumps(c2, sort_keys=True))
        self.assertEqual('{"a": 300}', json.dumps(c3, sort_keys=True))
        self.assertEqual('{"a": 400}', json.dumps(c4, sort_keys=True))


if __name__ == '__main__':
    unittest.main()



""" Test for json frontend """
import json
import shutil
import tempfile
import unittest

from pandas import DataFrame, Series

from filoc import FilocIO, filoc_pandas
# noinspection PyMissingOrEmptyDocstring
from filoc.contract import SingletonExpectedError, FrontendConversionError


# noinspection DuplicatedCode
# noinspection PyMissingOrEmptyDocstring
class TestPandasFrontend(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')
        self.path_fmt = self.test_dir + r'/simid={simid:d}/epid={epid:d}/hyperparameters.json'

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read_content(self):
        print("write files")
        wloc = FilocIO(self.path_fmt, writable=True)
        with wloc.open({"simid": 1, "epid": 10}, "w") as f:
            json.dump({'a': 100}, f)
        with wloc.open({"simid": 1, "epid": 20}, "w") as f:
            json.dump({'a': 200}, f)

        loc = filoc_pandas(self.path_fmt)

        # test 1
        # noinspection PyUnusedLocal
        p = loc.read_content(epid=10)

        # test 2
        try:
            # noinspection PyUnusedLocal
            p = loc.read_content()
            raise AssertionError("previous call expected to raise an error")
        except SingletonExpectedError:
            pass

    def test_write_contents_from_records(self):
        wloc = filoc_pandas(self.path_fmt, writable=True)
        # TODO: improve generic typing to avoid this line below
        # noinspection PyTypeChecker

        # ACT
        wloc.write_contents([
            {"simid": 1, "epid": 10, 'a': 100},
            {"simid": 1, "epid": 20, 'a': 200},
            {"simid": 2, "epid": 10, 'a': 300},
            {"simid": 2, "epid": 20, 'a': 400},
        ])

        # ASSERT
        wloc = FilocIO(self.path_fmt)
        with wloc.open({"simid": 1, "epid": 10}) as f: c1 = json.load(f)
        with wloc.open({"simid": 1, "epid": 20}) as f: c2 = json.load(f)
        with wloc.open({"simid": 2, "epid": 10}) as f: c3 = json.load(f)
        with wloc.open({"simid": 2, "epid": 20}) as f: c4 = json.load(f)

        self.assertEqual('{"a": 100}', json.dumps(c1, sort_keys=True))
        self.assertEqual('{"a": 200}', json.dumps(c2, sort_keys=True))
        self.assertEqual('{"a": 300}', json.dumps(c3, sort_keys=True))
        self.assertEqual('{"a": 400}', json.dumps(c4, sort_keys=True))

    def test_write_contents_from_DataFrame(self):
        wloc = filoc_pandas(self.path_fmt, writable=True)

        # ACT
        wloc.write_contents(DataFrame([
            {"simid": 1, "epid": 10, 'a': 100},
            {"simid": 1, "epid": 20, 'a': 200},
            {"simid": 2, "epid": 10, 'a': 300},
            {"simid": 2, "epid": 20, 'a': 400},
        ]))

        # ASSERT
        wloc = FilocIO(self.path_fmt)
        with wloc.open({"simid": 1, "epid": 10}) as f: c1 = json.load(f)
        with wloc.open({"simid": 1, "epid": 20}) as f: c2 = json.load(f)
        with wloc.open({"simid": 2, "epid": 10}) as f: c3 = json.load(f)
        with wloc.open({"simid": 2, "epid": 20}) as f: c4 = json.load(f)

        self.assertEqual('{"a": 100}', json.dumps(c1, sort_keys=True))
        self.assertEqual('{"a": 200}', json.dumps(c2, sort_keys=True))
        self.assertEqual('{"a": 300}', json.dumps(c3, sort_keys=True))
        self.assertEqual('{"a": 400}', json.dumps(c4, sort_keys=True))

    def test_write_contents_from_unsupported(self):
        wloc = filoc_pandas(self.path_fmt, writable=True)

        # ACT, ASSERT
        try:
            # noinspection PyTypeChecker
            wloc.write_contents(123)
            raise AssertionError("previous line must raise an exception")
        except FrontendConversionError:
            pass


    def test_write_content_from_dict(self):
        wloc = filoc_pandas(self.path_fmt, writable=True)
        # TODO: improve generic typing to avoid this line below
        # noinspection PyTypeChecker

        # ACT
        wloc.write_content({"simid": 1, "epid": 10, 'a': 100})

        # ASSERT
        wloc = FilocIO(self.path_fmt)
        with wloc.open({"simid": 1, "epid": 10}) as f: c1 = json.load(f)

        self.assertEqual('{"a": 100}', json.dumps(c1, sort_keys=True))

    def test_write_contents_from_Series(self):
        wloc = filoc_pandas(self.path_fmt, writable=True)

        # ACT
        wloc.write_content(Series({"simid": 1, "epid": 10, 'a': 100}))

        # ASSERT
        wloc = FilocIO(self.path_fmt)
        with wloc.open({"simid": 1, "epid": 10}) as f: c1 = json.load(f)

        self.assertEqual('{"a": 100}', json.dumps(c1, sort_keys=True))

    def test_write_content_from_unsupported(self):
        wloc = filoc_pandas(self.path_fmt, writable=True)

        # ACT, ASSERT
        try:
            # noinspection PyTypeChecker
            wloc.write_content(['AHHHHH'])
            raise AssertionError("previous line must raise an exception")
        except FrontendConversionError:
            pass


if __name__ == '__main__':
    unittest.main()



import os
import shutil
import tempfile
import unittest
from io import UnsupportedOperation
from pathlib import Path

from filoc.filoc_io import FilocIO


def touch_file(file_path):
    os.makedirs(os.path.dirname(file_path))
    Path(file_path).touch()


# noinspection DuplicatedCode
# noinspection PyMissingOrEmptyDocstring
class TestFilocIO(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')
        self.path_fmt = self.test_dir + r'/simid={simid:d}/epid={epid:d}/hyperparameters.json'

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_get_path_properties(self):
        loc = FilocIO(self.path_fmt)
        props = loc.parse_path_properties(rf"{self.test_dir}/simid=12/epid=102/hyperparameters.json")
        self.assertEqual(len(props), 2)
        self.assertEqual(props['simid'], 12)
        self.assertEqual(props['epid'], 102)

    def test_get_path(self):
        loc = FilocIO(self.path_fmt)
        path1 = loc.render_path(simid=12, epid=102)
        self.assertEqual(path1, rf"{self.test_dir}/simid=12/epid=102/hyperparameters.json")
        # Todo: test other formattings: float, string

    def test_get_glob_path(self):
        loc = FilocIO(self.path_fmt)
        path1 = loc.render_glob_path(epid=102)
        self.assertEqual(path1, rf"{self.test_dir}/simid=?*/epid=102/hyperparameters.json")
        # Todo: test other formattings: float, string

    def test_find_paths(self):
        loc = FilocIO(self.path_fmt)
        touch_file(loc.render_path(simid=1, epid=10))
        touch_file(loc.render_path(simid=1, epid=20))
        touch_file(loc.render_path(simid=2, epid=10))
        touch_file(loc.render_path(simid=2, epid=20))

        p = loc.list_paths(simid=1)
        self.assertListEqual(p, [
            rf"{self.test_dir}/simid=1/epid=10/hyperparameters.json",
            rf"{self.test_dir}/simid=1/epid=20/hyperparameters.json"
        ])

        p = loc.list_paths(epid=10)
        self.assertListEqual(p, [
            rf"{self.test_dir}/simid=1/epid=10/hyperparameters.json",
            rf"{self.test_dir}/simid=2/epid=10/hyperparameters.json"
        ])

        p = loc.list_paths(epid=12)
        self.assertListEqual(p, [])

    def test_find_paths_and_path_props(self):
        loc = FilocIO(self.path_fmt)
        touch_file(loc.render_path(simid=1, epid=10))
        touch_file(loc.render_path(simid=1, epid=20))
        touch_file(loc.render_path(simid=2, epid=10))
        touch_file(loc.render_path(simid=2, epid=20))

        p = loc.list_paths_and_props(simid=1)
        self.assertListEqual(p, [
            (rf"{self.test_dir}/simid=1/epid=10/hyperparameters.json", {'simid': 1, 'epid': 10}),
            (rf"{self.test_dir}/simid=1/epid=20/hyperparameters.json", {'simid': 1, 'epid': 20}),
        ])

        p = loc.list_paths_and_props(epid=10)
        self.assertListEqual(p, [
            (rf"{self.test_dir}/simid=1/epid=10/hyperparameters.json", {'simid': 1, 'epid': 10}),
            (rf"{self.test_dir}/simid=2/epid=10/hyperparameters.json", {'simid': 2, 'epid': 10}),
        ])

        p = loc.list_paths_and_props(epid=12)
        self.assertListEqual(p, [])

    def test_exists(self):
        loc = FilocIO(self.path_fmt)
        self.assertEqual(loc.exists(simid=1, epid=10), False)
        touch_file(loc.render_path(simid=1, epid=10))
        self.assertEqual(loc.exists(simid=1, epid=10), True)

    def test_open_write_readonly(self):
        loc = FilocIO(self.path_fmt)
        with self.assertRaises(UnsupportedOperation):
            with loc.open(dict(simid=1, epid=10), 'w') as f:
                f.write("coucou")

    def test_open_write_read(self):
        loc = FilocIO(self.path_fmt, writable=True)
        with loc.open(dict(simid=1, epid=10), 'w') as f:
            f.write("coucou")
        with loc.open(dict(simid=1, epid=10), 'r') as f:
            s = f.read()
        self.assertEqual(s, "coucou")

    def test_delete_readonly(self):
        loc = FilocIO(self.path_fmt)
        touch_file(loc.render_path(simid=1, epid=10))
        with self.assertRaises(UnsupportedOperation):
            loc.delete(dict(simid=1, epid=10))

    def test_delete(self):
        loc = FilocIO(self.path_fmt, writable=True)
        touch_file(loc.render_path(simid=1, epid=10))
        self.assertEqual(loc.exists(simid=1, epid=10), True)
        loc.delete(dict(simid=1, epid=10))
        self.assertEqual(loc.exists(simid=1, epid=10), False)


if __name__ == '__main__':
    unittest.main()



""" This module contains the filoc default frontend implementations """
from .frontend_json import JsonFrontend
from .frontend_pandas import PandasFrontend

__all__     = [
    'JsonFrontend',
    'PandasFrontend',
]



"""
Filoc public API
"""
from .contract import Filoc
from .core import FilocSingle, FilocComposite
from .filoc_io import FilocIO
from filoc.factories import filoc, filoc_json, filoc_pandas

__version__ = '0.0.7'


__all__     = [
    'filoc',
    'filoc_json', 
    'filoc_pandas'
    'FilocIO',
    'Filoc',
    'FilocSingle',
    'FilocComposite',
]



"""
This module contains the contracts of filoc classes. Among other it contains the ``FrontendContract`` and``BackendContract``, which
custom frontends and backends need to implement.
"""

# -------------
# Types aliases
# -------------
import logging
from abc import ABC, abstractmethod
from typing import TypeVar, Any, List, Generic, Optional, Mapping, Dict, Collection
from fsspec import AbstractFileSystem

# Literal is python 3.8 feature, but filoc works from python 3.6 upward
try:
    from typing import Literal
except ImportError:
    Literal = None

# set logging level to WARN. User can override the level to info, to get more infos
logging.getLogger('filoc').setLevel(logging.WARN)

TContent             = TypeVar('TContent')
"""Generic type of objects returned by ``Filoc.get_content(...)`` and expected by ``Filoc.write_content(...)``. 
For example, in the ``'json'`` frontend, TContent is equal to `Mapping[str,Any]``, and in the ``'pandas'`` frontend, TContent is equal to ``pandas.Series`` """

TContents            = TypeVar('TContents')
"""Generic type of objects returned by ``Filoc.get_contents(...)`` and expected by ``Filoc.write_contents(...)``. 
For example, in the ``'json'`` frontend, TContents is equal to `List[Mapping[str,Any]]``, and in the ``'pandas'`` frontend, TContent is equal to ``pandas.DataFrame`` """

BuiltinFrontends      = Literal['json', 'pandas'] if Literal else str
"""Shortcuts used to designate filoc preset frontends"""

BuiltinBackends       = Literal['json', 'yaml', 'csv', 'pickle'] if Literal else str
"""Shortcut used to designate filoc preset backends"""

Constraint           = Any
"""key-values describing constraints to apply while filtering data. Currently only equality is supported."""

Constraints          = Mapping[str, Constraint]
"""key-values describing constraints to apply while filtering data. Currently only equality is supported."""

PropValue            = Any

ReadOnlyProps        = Mapping[str, PropValue]
"""filoc intermediate data representation of a single data 'row'"""

Props                = Dict[str, PropValue]
"""filoc intermediate data representation of a single data 'row'"""

ReadOnlyPropsList    = Collection[ReadOnlyProps]
"""filoc intermediate data representation between the backend files and the frontend TContent and TContents objects"""

PropsList            = List[Props]
"""filoc intermediate data representation between the backend files and the frontend TContent and TContents objects"""


# -----------------
# Exception classes
# -----------------
class FrontendConversionError(ValueError):
    """Exception raised by filoc frontends, when the frontend encounters a conversion problem """
    def __init__(self, *args):
        super().__init__(*args)


class SingletonExpectedError(FrontendConversionError):
    """Exception raised by filoc frontends, when the frontend expects a single entry but got multiple entries """
    def __init__(self, *args):
        super().__init__(*args)

# ----------------
# Abstract classes
# ----------------
# TODO: Unit tests with locpath = folders instead of path


class BackendContract(ABC):
    """The abstract class that filoc backends need to implement."""

    @abstractmethod
    def read(self, fs: AbstractFileSystem, path: str, path_props : Props, constraints: Constraints) -> PropsList:
        """Reads the data contained at ``path`` on the file system ``fs``, applies additional filters defined in ``constraints`` and convert the data to the filoc intermediate representation.

        Args:
            fs: File system implementation (see the `fsspec <https://github.com/intake/filesystem_spec/>`_ library)
            path: The path at which the data must be loaded from. It is a concrete form of the filoc ``locpath``.
            path_props: the key-values extracted from the path
            constraints: All constraints provided by the user to take into account.

        Returns:
            PropsList: filoc intermediate representation between frontend and backend data (list of dictionary).
        """
        raise NotImplementedError("Abstract")

    @abstractmethod
    def write(self, fs : AbstractFileSystem, path : str, props_list : PropsList) -> None:
        """Writes the data contained in ``props_list`` into ``path`` on the file system ``fs``.

        Args:
            fs: File system implementation (see the `fsspec <https://github.com/intake/filesystem_spec/>`_ library)
            path: The path to which the data must be saved to. It is a concrete form of the filoc ``locpath``. 
            props_list: filoc intermediate representation between frontend and backend data (list of dictionary).
        """
        raise NotImplementedError("Abstract")


class FrontendContract(Generic[TContent, TContents], ABC):
    """The abstract class that filoc frontends need to implement.

    TContent ([Any]): The type returned by ``get_content(...)`` and expected by ``write_content(...)``
    TContents ([Any]): The type returned by ``get_contents(...)`` and expected by ``write_contents(...)``
    """

    @abstractmethod
    def read_content(self, props_list : PropsList) -> TContent:
        """ Converts the ``props_list`` intermediate representation into a frontend ``TContent`` object. 
        
        Implementation remark: As you implement this method, You may want to first validate that the ``props_list``
        contains a single item. But you may want also to support multiple items, for example if you need to aggregate multiple files at folder level. 

        Args:
            props_list: filoc intermediate representation between frontend and backend data (list of dictionary).

        Returns:
            TContent: The frontend content object (i.e. the default json frontend implementation returns a json object)
        """
        raise NotImplementedError("Abstract")

    @abstractmethod
    def read_contents(self, props_list : PropsList) -> TContents:
        """ Converts the ``props_list`` intermediate representation into a frontend ``TContents`` object. 

        Args:
            props_list: filoc intermediate representation between frontend and backend data (list of dictionary).

        Returns:
            TContents: The frontend contents object (i.e. the default json frontend implementation returns a json list)
        """
        raise NotImplementedError("Abstract")

    @abstractmethod
    def write_content(self, content : TContent) -> ReadOnlyPropsList:
        """ Converts the frontend ``TContent`` object into the filoc intermediate representation.

        Args:
            content: The frontend content object (i.e. the default json frontend implementation returns a json object)

        Returns:
            PropsList: filoc intermediate representation between frontend and backend data (list of dictionary).
        """
        raise NotImplementedError("Abstract")

    @abstractmethod
    def write_contents(self, contents : TContents) -> ReadOnlyPropsList:
        """ Converts the frontend ``TContents`` object into the filoc intermediate representation.

        Args:
            contents: The frontend contents object (i.e. the default json frontend implementation returns a json list)

        Returns:
            PropsList: filoc intermediate representation between frontend and backend data (list of dictionary).
        """
        raise NotImplementedError("Abstract")


class Filoc(Generic[TContent, TContents], ABC):
    """ The contract of objects created by the filoc factory. This is the most important contract in the filoc library.

    TContent ([Any]): The type returned by ``get_content(...)`` and expected by ``write_content(...)``
    TContents ([Any]): The type returned by ``get_contents(...)`` and expected by ``write_contents(...)``
    """

    @abstractmethod
    def lock(self, attempt_count: int = 60, attempt_secs: float = 1.0):
        """Prevents other filoc instances to concurrently read or write any file in the filoc tree. Usage:

        .. code-block:: python

            with my_filoc.lock():
                my_filoc.read_contents(df)
                my_filoc.write_contents(df)

        Raises:
            NotImplementedError: [description]
        """
        raise NotImplementedError('Abstract')

    @abstractmethod
    def lock_info(self) -> Optional[Mapping[str, Any]]:
        """Returns the information contained in the lock file(s) augmented of the file timestamp if it exists, elsewhere null.
        For a single filoc, it returns simple key-values. 
        For nested filocs (composite), it returns the lock file information in the same tree structure as the filoc definition tree. 
        
        This function is useful in case of problems, in order to analyze what's happening (mostly to check whether the lock owner is still alive)"""
        raise NotImplementedError('Abstract')

    @abstractmethod
    def lock_force_release(self):
        """Forces the removing of lock file even a concurrent process is still supposed to own it. Consider this method as your "last action before being fired"!"""
        raise NotImplementedError('Abstract')

    @abstractmethod
    def invalidate_cache(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint):
        """Delete the cache data for the provided constraints, to ensure that the data will be re-fetched by the filoc backend at the next ``get_content(...)`` or ``get_contents(...)`` calls.
        
        Remark: By default, the cache is configured to automatically invalidate cache entries loaded from paths whose timestamp changed. All default backends work on files and this
        implementation works. But for custom backends, this invalidation mechanism may not work anymore (for example if the backend allows to work with folders instead of files). 
        In that case, it may be required to invalidate the cache manually with the current method. """
        raise NotImplementedError('Abstract')

    @abstractmethod
    def read_content(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> TContent:
        """Reads the data entries fulfilling the ``constraints`` and converts them to a TContent object. With the filoc default frontend implementations, the ``constraints`` must result in the
        selection of a single entry. If multiple rows are selected, the call raises a ``SingletonExpectedError``

        Args:
            constraints: Filter criteria used to shrink the result set. Defaults to None.
            **constraints_kwargs: additional constraints merge together with ``constraints`` argument
        Returns:
            TContent: An instance of the frontend type for a single entry (ex: ``json`` frontend: Mapping[str, Any], ``pandas`` frontend: pandas.Series)

        Raises:
            SingletonExpectedError: if more than one entry is selected after application of the ``constraints`` (behavior of filoc default frontends)
            FrontendConversionError: if the frontend cannot perform the conversion 
        """
        raise NotImplementedError('Abstract')

    @abstractmethod
    def read_contents(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> TContents:
        """Reads the data entries fulfilling the ``constraints`` and converts them to a TContents object. 

        Args:
            constraints: Filter criteria used to shrink the result set. Defaults to None.
            **constraints_kwargs: additional constraints merge together with ``constraints`` argument
        Returns:
            TContents: An instance of the frontend types for multiple entries (ex: ``json`` frontend: List[Mapping[str, Any]], ``pandas`` frontend: pandas.DataFrame)

        Raises:
            FrontendConversionError: If the frontend cannot perform the conversion 
        """
        raise NotImplementedError('Abstract')

    @abstractmethod
    def _read_props_list(self, constraints : Optional[Constraints] = None, **constraints_kwargs : Constraint) -> PropsList:
        """Reads the data entries fulfilling the ``constraints`` and converts them to the filoc intermediate representation.
        This function is used internally to shared the intermediate representation in filoc composites.

        Args:
            constraints: Filter criteria used to shrink the result set. Defaults to None.
            **constraints_kwargs: additional constraints merge together with ``constraints`` argument
        Returns:
            PropsList: The filoc intermediate representation
        """
        raise NotImplementedError('Abstract')

    @abstractmethod
    def write_content(self, content : TContent, dry_run=False):
        """Writes the ``content`` entry to the file system.

        Args:
            content: The content accepted by the frontend (ex: ``json`` frontend: Mapping[str, Any], ``pandas`` frontend: pandas.Series or Mapping[str, Any])
            dry_run: If True, then only simulate the writing. The default value is False.

        Raises:
            FrontendConversionError: If the frontend cannot perform the conversion 
        """
        raise NotImplementedError('Abstract')

    @abstractmethod
    def write_contents(self, contents : TContents, dry_run=False):
        """Writes the ``contents`` entries to the file system.

        Args:
            contents: The content accepted by the frontend (ex: ``json`` frontend: List[Mapping[str, Any]], ``pandas`` frontend: pandas.DataFrame or List[Mapping[str, Any]])
            dry_run: If True, then only simulate the writing. The default value is False.

        Raises:
            FrontendConversionError: If the frontend cannot perform the conversion 
        """
        raise NotImplementedError('Abstract')

    @abstractmethod
    def _write_props_list(self, props_list : PropsList, dry_run=False):
        """Write the filoc intermediate representation to the file system via the backend.
        This function is used internally to shared the intermediate representation in filoc composites.

        Args:
            props_list: the filoc intermediate representation 
            dry_run: If True, then only simulate the writing. The default value is False.
        """
        raise NotImplementedError('Abstract')



""" github protocol test """
import os
import shutil
import tempfile
import unittest
from pathlib import Path

from filoc import filoc


# noinspection PyMissingOrEmptyDocstring
def touch_file(file_path):
    os.makedirs(os.path.dirname(file_path))
    Path(file_path).touch()


# noinspection PyMissingOrEmptyDocstring,PyPep8Naming
class Test_fs_github(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read_contents(self):
        loc = filoc('github://CSSEGISandData:COVID-19@/csse_covid_19_data/csse_covid_19_daily_reports/{date_str}.csv', backend='csv', singleton=False)
        loc.read_contents(date_str='01-27-2020')



import json
import shutil
import tempfile
import unittest

# noinspection DuplicatedCode
from filoc import filoc_json, FilocIO


# noinspection PyMissingOrEmptyDocstring
class TestMultiloc(unittest.TestCase):
    def setUp(self):
        self.maxDiff               = None
        self.test_dir              = tempfile.mkdtemp().replace('\\', '/')
        self.path_fmt_simid_config = self.test_dir + r'/somewhere1/simid={simid:d}/config.json'
        self.path_fmt_hyp          = self.test_dir + r'/somewhere1/simid={simid:d}/epid={epid:d}/hyperparameters.json'
        self.path_fmt_res          = self.test_dir + r'/somewhere1/epid={epid:d}/simid={simid:d}/result.json'
        self.conf_wloc             = filoc_json(self.path_fmt_simid_config, writable=True )
        self.conf_loc              = filoc_json(self.path_fmt_simid_config, writable=False)
        self.hyp_loc               = filoc_json(self.path_fmt_hyp         , writable=True )
        self.res_loc               = filoc_json(self.path_fmt_res         , writable=True )

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read_contents(self):
        pass

    def test_write_and_read_contents(self):
        # ACT 1 (write)
        self.conf_wloc.write_contents([
            {"simid": 1, "confA" : "Q"},
            {"simid": 2, "confA" : "R"},
        ])

        mloc = filoc_json({'conf' : self.conf_loc, 'hyp' : self.hyp_loc, 'res' : self.res_loc})
        mloc.write_contents([
            {"shared.simid": 1, "shared.epid": 10, "hyp.a": 100, "res.b": 1000},
            {"shared.simid": 1, "shared.epid": 20, "hyp.a": 200, "res.b": 2000},
            {"shared.simid": 2, "shared.epid": 10, "hyp.a": 300, "res.b": 3000},
            {"shared.simid": 2, "shared.epid": 20, "hyp.a": 400, "res.b": 4000},
        ])

        wloc = FilocIO(self.path_fmt_hyp)
        with wloc.open({"simid": 1, "epid": 10}) as f: hyp1 = json.load(f)
        with wloc.open({"simid": 1, "epid": 20}) as f: hyp2 = json.load(f)
        with wloc.open({"simid": 2, "epid": 10}) as f: hyp3 = json.load(f)
        with wloc.open({"simid": 2, "epid": 20}) as f: hyp4 = json.load(f)

        self.assertEqual('{"a": 100}', json.dumps(hyp1, sort_keys=True))
        self.assertEqual('{"a": 200}', json.dumps(hyp2, sort_keys=True))
        self.assertEqual('{"a": 300}', json.dumps(hyp3, sort_keys=True))
        self.assertEqual('{"a": 400}', json.dumps(hyp4, sort_keys=True))

        wloc = FilocIO(self.path_fmt_res)
        with wloc.open({"simid": 1, "epid": 10}) as f: hyp1 = json.load(f)
        with wloc.open({"simid": 1, "epid": 20}) as f: hyp2 = json.load(f)
        with wloc.open({"simid": 2, "epid": 10}) as f: hyp3 = json.load(f)
        with wloc.open({"simid": 2, "epid": 20}) as f: hyp4 = json.load(f)

        self.assertEqual('{"b": 1000}', json.dumps(hyp1, sort_keys=True))
        self.assertEqual('{"b": 2000}', json.dumps(hyp2, sort_keys=True))
        self.assertEqual('{"b": 3000}', json.dumps(hyp3, sort_keys=True))
        self.assertEqual('{"b": 4000}', json.dumps(hyp4, sort_keys=True))

        # ACT 2 (read)
        p = mloc.read_contents(epid=10)
        self.assertEqual(len(p), 2)

        self.assertEqual(
            '[{"conf.confA": "Q", "hyp.a": 100, "res.b": 1000, "shared.epid": 10, "shared.simid": 1}, {"conf.confA": "R", "hyp.a": 300, "res.b": 3000, "shared.epid": 10, "shared.simid": 2}]',
            json.dumps(p, sort_keys=True))

        # ACT 3 (update)
        p[1]['res.b'] = 3333
        mloc.write_contents(p)
        
        self.assertEqual(
            '[{"conf.confA": "Q", "hyp.a": 100, "res.b": 1000, "shared.epid": 10, "shared.simid": 1}, {"conf.confA": "R", "hyp.a": 300, "res.b": 3333, "shared.epid": 10, "shared.simid": 2}]',
            json.dumps(p, sort_keys=True))

        # ACT 3 (add attribute)
        p[1]['res.c'] = 'NEW'
        mloc.write_contents(p)
        
        self.assertEqual(
            '[{"conf.confA": "Q", "hyp.a": 100, "res.b": 1000, "shared.epid": 10, "shared.simid": 1}, {"conf.confA": "R", "hyp.a": 300, "res.b": 3333, "res.c": "NEW", "shared.epid": 10, "shared.simid": 2}]',
            json.dumps(p, sort_keys=True))


if __name__ == '__main__':
    unittest.main()



""" Filoc default pickle backend implementation """
import os
import pickle
from typing import Dict, Any

from fsspec import AbstractFileSystem

from filoc.contract import PropsList, BackendContract, Constraints, Props
from filoc.utils import filter_and_coerce_loaded_file_content, coerce_file_content_to_write


class PickleBackend(BackendContract):
    """
    filoc backend used to read data from Pickle files and write into them. This implementation is used when you call the filoc factory with the ``backend`` argument set to ``'pickle'``. Example:
    
    .. code-block:: python

        loc = filoc('/my/locpath/{id}/data.pickle', backend='pickle')

    It is recommended to read files that you wrote with filoc itself. If you want to read pickle files written by a third library, it is recommended to implement your own backend, 
    so that you can better handle the edge cases and print out better error messages.
    """
    def __init__(self, is_singleton) -> None:
        super().__init__()
        self.is_singleton = is_singleton

    def read(self, fs: AbstractFileSystem, path: str, path_props : Props, constraints: Constraints) -> PropsList:
        """(see BackendContract contract) """
        with fs.open(path, 'rb') as f:
            return filter_and_coerce_loaded_file_content(path, pickle.load(f), path_props, constraints, self.is_singleton)

    def write(self, fs: AbstractFileSystem, path: str, props_list: PropsList) -> None:
        """(see BackendContract contract)"""
        fs.makedirs(os.path.dirname(path), exist_ok=True)
        with fs.open(path, 'wb') as f:
            return pickle.dump(coerce_file_content_to_write(path, props_list, self.is_singleton), f)



""" Test """
import os
import shutil
import socket
import tempfile
import threading
import time
import unittest
from threading import Thread

from filoc import filoc
from filoc.core import LockException


# noinspection DuplicatedCode,PyMissingOrEmptyDocstring
class TestFilocConcurrency(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')
        self.loc = filoc(self.test_dir + r'/id={id:d}/myfile.json', writable=True)

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_lock_block(self):
        with self.loc.lock():
            self.loc.write_content({'id' : 1, 'val' : 1})
            res = self.loc.read_content(id=1)
            print(res)

    def test_lock_info(self):
        host   = socket.gethostname()
        pid    = os.getpid()
        thread = threading.get_ident()

        self.assertIsNone(self.loc.lock_info())

        with self.loc.lock():            
            lock_info = self.loc.lock_info()
            
            self.assertEqual(lock_info['host'], host)
            self.assertEqual(lock_info['pid'], pid)
            self.assertEqual(lock_info['thread'], thread)
            
            self.loc.write_content({'id' : 1, 'val' : 1})
            res = self.loc.read_content(id=1)
            print(res)

        self.assertIsNone(self.loc.lock_info())

    def test_lock_force(self):
        self.loc.lock().__enter__()            
        self.loc.lock_force_release()
        self.assertIsNone(self.loc.lock_info())

    def test_lock_force2(self):
        with self.loc.lock():
            self.loc.lock_force_release()
            self.assertIsNone(self.loc.lock_info())

    def test_lock_reenter(self):
        with self.loc.lock():
            with self.loc.lock():
                pass

    def test_lock_enter_from_other_thread(self):

        state = 0

        def wait_state_and_increment(expected_state):
            nonlocal state
            while state != expected_state:
                time.sleep(0.1)
            state += 1

        def async_lock():
            with self.loc.lock():
                wait_state_and_increment(1)
                wait_state_and_increment(4)
                    
        thread = Thread(target=async_lock)

        # BEGIN ASYNC PLAY
        wait_state_and_increment(0)  # state 0 --> 1
        thread.start()               # state 1 --> 2 (then lock is set)
        wait_state_and_increment(2)  # state 2 --> 3

        try:
            with self.loc.lock(attempt_count=3, attempt_secs=0.2):
                self.fail("this line should never be called")
        except LockException:
            print("lock worked")
        finally:
            wait_state_and_increment(3)  # state 3 --> 4 (trigger lock release)
            wait_state_and_increment(5)  # state 5 --> 6 (lock released)


if __name__ == '__main__':
    unittest.main()



"""
This module contains the filoc factories ``filoc(...)``, ``filoc_json(...)``, ``filoc_pandas(...)``.
"""
from typing import Any, Dict, List, Mapping, Optional, Union, overload

from fsspec.spec import AbstractFileSystem
from pandas.core.frame import DataFrame
from pandas.core.series import Series

from filoc.contract import BuiltinFrontends, Filoc, BackendContract, FrontendContract, BuiltinBackends
from filoc.core import FilocSingle, FilocComposite

_default_frontend        = 'pandas'
_default_backend         = 'json'
_default_singleton       = True
_default_encoding        = None
_default_writable        = False
_default_transaction     = True
_default_timestamp_col   = None
_default_join_level_name = 'shared'
_default_join_separator  = '.'


def _get_frontend(frontend : Union[BuiltinFrontends, FrontendContract]):
    if frontend == 'json':
        from filoc.frontends import JsonFrontend
        return JsonFrontend()
    elif frontend == 'pandas':
        from filoc.frontends import PandasFrontend
        return PandasFrontend()
    else:
        return frontend


def _get_backend(backend : Union[BuiltinBackends, BackendContract], is_singleton : bool, encoding : str):
    if backend == 'csv':
        from filoc.backends import CsvBackend
        return CsvBackend(encoding)
    elif backend == 'json':
        from filoc.backends import JsonBackend
        return JsonBackend(is_singleton, encoding)
    elif backend == 'pickle':
        from filoc.backends import PickleBackend
        return PickleBackend(is_singleton)
    elif backend == 'yaml':
        from filoc.backends import YamlBackend
        return YamlBackend(is_singleton, encoding)
    else:
        return backend

@overload
def filoc(
        locpath          : str,
        frontend         : Union[BuiltinFrontends, FrontendContract] = _default_frontend,
        backend          : Union[BuiltinBackends, BackendContract]   = _default_backend,
        singleton        : bool                                      = _default_singleton,
        encoding         : Optional[str]                             = _default_encoding,
        writable         : Optional[bool]                            = _default_writable,
        transaction      : Optional[bool]                            = _default_transaction,
        cache_locpath    : Optional[str]                             = None,
        cache_fs         : Optional[AbstractFileSystem]              = None,
        timestamp_col    : Optional[str]                             = _default_timestamp_col,
        join_level_name  : Optional[str]                             = _default_join_level_name,
        join_separator   : Optional[str]                             = _default_join_separator,
        fs               : Optional[AbstractFileSystem]              = None,
) -> FilocSingle:
    """
    Creates a ``Filoc`` instance which allows to read a *set of files* and visualize it as a DataFrame (or another frontend object, if another frontend is passed to the factory), and write changes back to thhe *set of files*.

    Args:
        locpath: 
            A path which can be either a local path (``/data/file.json``) or a fsspec path with protocol (i.e. ``ftp://user:pass@example.com/data/file.yaml``). The ``locpath``
            may additionally contain format placeholders (``/data/{customer_id}/info.json``). during reading accesses, the placeholder names and their values in the concrete paths
            are merged with the data loaded from the file contents. On writing, the placeholders are substituted by the related values found in the frontend object (i.e. DataFrame).
            
            The locpath can also be a Mapping (dictionary). In that case, ``filoc(...)`` creates a "composite filoc" built of multiple sub-filocs. The key of a mapping entry is 
            the name of the sub-filoc and the value a ``Filoc`` instance. You can alternatively provide a string, which is used as ``locpath`` in a recursive call to ``filoc(...)`` 
            in order to instantiate the sub-filoc.

            Example 1: ``/data/{country}/{company}/info.json``

            Example 2: ``{ 'user' : '/data/user_data/{user_id:d}/user_info.json', 'activity' : already_instantiated_filoc }``

        frontend: 
            The frontend determines the types of objects returned by ``Filoc.read_contents(...)``, ``Filoc.read_content(...)`` and accepted by  
            ``Filoc.write_contents(...)``, ``Filoc.write_content(...)``. The default frontend is ``'pandas'``. The two builtin frontends are ``'pandas'`` and ``'json'``. 
            You can also provide your own frontend instance, which must implement the ``FrontendContract``.

        backend:
            The backend determines how Filoc reads and writes the files. The default backend is ``json``. The four builtin backends are 
            ``json``, ``yaml``, ``csv`` and ``pickle``. You can also provide your own backend instance, which must implement the ``BackendContract``.

        singleton: 
            Option that applies only to the three builtin backends: ``json``, ``yaml`` and ``pickle``. The default value is ``True``. If ``True``, then the created ``Filoc`` instance assumes that the
            files contain a single Mapping (in JSON: a JSON object). If ``False``, the Filoc assumes that the files contains a list of Mappings (in JSON: a JSON array of JSON objects).

        encoding:
            Option that applies only to the three builtin backends: ``json``, ``yaml`` and ``csv``. The default value is ``None``, keeping the default value of the underlying fsspec file system. 
            Allows to override the file encoding.

        writable:
            If True, then write operations are allowed. Elsewhere write operation raises an ``UnsupportedOperation`` error.

        transaction:
            If True, then write into a temp file and then rename the file to the target file.
            Renaming is on certain file system is atomic, which provide a cheap transaction behavior on a single file.

        cache_locpath:
            Default: None. Defines a locpath string (fsspec path with format placeholders), used by the created ``Filoc`` instance to cache 
            the intermediate representation (data model loaded by the backend). The cache allows to speed up multiple reads from slow remote
            file systems as well as speed up expensive backend processing. 
            
            If you provide a ``cache_locpath`` with placeholders, the cache is then splitted into multiple files based on the placeholder values. 
            It allows you to encapsulate the cache data with original the data structure.

        cache_fs:
            Default: None. Allows to provide a custom instance of fsspec file system for the cache files defined by ``cache_locpath``. This may be 
            required, if the protocol used in ``cache_locpath`` needs to be configured or fine-tuned (ex: ``ftp://``).

        timestamp_col:
            Default: None. If not None, then filoc adds a column ``timestamp_col``, which contains the timestamp of the file the data entries comes from.

        join_level_name:
            Default: ``'shared'``. This option applies only to composite filocs and allows to provide an alternative prefix to the "index columns", the columns that 
            are used to join the sub-filocs together.

        join_separator:
            Default: ``'.'``. This option applies only to composite filocs and allows to provide an alternative separator between the "sub-filoc" name 
            and the attributes names of the sub-filoc.

        fs:
            Default: None. Allows to provide a custom instance of fsspec file system. This may be required, if the protocol used in the ``locpath`` needs to be 
            configured or fine-tuned (ex: ``ftp://``).

    Returns:
        A ``Filoc`` instance
    """
    ...

@overload
def filoc(
        locpath          : Mapping[str, Union[str, Filoc]],
        frontend         : Union[BuiltinFrontends, FrontendContract] = _default_frontend,
        backend          : Union[BuiltinBackends, BackendContract]   = _default_backend,
        singleton        : bool                                      = _default_singleton,
        encoding         : Optional[str]                             = _default_encoding,
        writable         : Optional[bool]                            = _default_writable,
        transaction      : Optional[bool]                            = _default_transaction,
        cache_locpath    : Optional[str]                             = None,
        cache_fs         : Optional[AbstractFileSystem]              = None,
        timestamp_col    : Optional[str]                             = _default_timestamp_col,
        join_level_name  : Optional[str]                             = _default_join_level_name,
        join_separator   : Optional[str]                             = _default_join_separator,
        fs               : Optional[AbstractFileSystem]              = None,
) -> FilocComposite:
    """
    Creates a ``Filoc`` instance which allows to read a *set of files* and visualize it as a DataFrame (or another frontend object, if another frontend is passed to the factory), and write changes back to thhe *set of files*.

    Args:
        locpath: 
            A path which can be either a local path (``/data/file.json``) or a fsspec path with protocol (i.e. ``ftp://user:pass@example.com/data/file.yaml``). The ``locpath``
            may additionally contain format placeholders (``/data/{customer_id}/info.json``). during reading accesses, the placeholder names and their values in the concrete paths
            are merged with the data loaded from the file contents. On writing, the placeholders are substituted by the related values found in the frontend object (i.e. DataFrame).
            
            The locpath can also be a Mapping (dictionary). In that case, ``filoc(...)`` creates a "composite filoc" built of multiple sub-filocs. The key of a mapping entry is 
            the name of the sub-filoc and the value a ``Filoc`` instance. You can alternatively provide a string, which is used as ``locpath`` in a recursive call to ``filoc(...)`` 
            in order to instantiate the sub-filoc.

            Example 1: ``/data/{country}/{company}/info.json``

            Example 2: ``{ 'user' : '/data/user_data/{user_id:d}/user_info.json', 'activity' : already_instantiated_filoc }``

        frontend: 
            The frontend determines the types of objects returned by ``Filoc.read_contents(...)``, ``Filoc.read_content(...)`` and accepted by  
            ``Filoc.write_contents(...)``, ``Filoc.write_content(...)``. The default frontend is ``'pandas'``. The two builtin frontends are ``'pandas'`` and ``'json'``. 
            You can also provide your own frontend instance, which must implement the ``FrontendContract``.

        backend:
            The backend determines how Filoc reads and writes the files. The default backend is ``json``. The four builtin backends are 
            ``json``, ``yaml``, ``csv`` and ``pickle``. You can also provide your own backend instance, which must implement the ``BackendContract``.

        singleton: 
            Option that applies only to the three builtin backends: ``json``, ``yaml`` and ``pickle``. The default value is ``True``. If ``True``, then the created ``Filoc`` instance assumes that the
            files contain a single Mapping (in JSON: a JSON object). If ``False``, the Filoc assumes that the files contains a list of Mappings (in JSON: a JSON array of JSON objects).

        encoding:
            Option that applies only to the three builtin backends: ``json``, ``yaml`` and ``csv``. The default value is ``None``, keeping the default value of the underlying fsspec file system. 
            Allows to override the file encoding.

        writable:
            If True, then write operations are allowed. Elsewhere write operation raises an ``UnsupportedOperation`` error.

        transaction:
            If True, then write into a temp file and then rename the file to the target file.
            Renaming is on certain file system is atomic, which provide a cheap transaction behavior on a single file.

        cache_locpath:
            Default: None. Defines a locpath string (fsspec path with format placeholders), used by the created ``Filoc`` instance to cache 
            the intermediate representation (data model loaded by the backend). The cache allows to speed up multiple reads from slow remote
            file systems as well as speed up expensive backend processing. 
            
            If you provide a ``cache_locpath`` with placeholders, the cache is then splitted into multiple files based on the placeholder values. 
            It allows you to encapsulate the cache data with original the data structure.

        cache_fs:
            Default: None. Allows to provide a custom instance of fsspec file system for the cache files defined by ``cache_locpath``. This may be 
            required, if the protocol used in ``cache_locpath`` needs to be configured or fine-tuned (ex: ``ftp://``).

        timestamp_col:
            Default: None. If not None, then filoc adds a column ``timestamp_col``, which contains the timestamp of the file the data entries comes from.

        join_level_name:
            Default: ``'shared'``. This option applies only to composite filocs and allows to provide an alternative prefix to the "index columns", the columns that 
            are used to join the sub-filocs together.

        join_separator:
            Default: ``'.'``. This option applies only to composite filocs and allows to provide an alternative separator between the "sub-filoc" name 
            and the attributes names of the sub-filoc.

        fs:
            Default: None. Allows to provide a custom instance of fsspec file system. This may be required, if the protocol used in the ``locpath`` needs to be 
            configured or fine-tuned (ex: ``ftp://``).

    Returns:
        A ``Filoc`` instance
    """
    ...

def filoc(
        locpath          : Union[str, Mapping[str, Union[str, Filoc]]],
        frontend         : Union[BuiltinFrontends, FrontendContract] = _default_frontend,
        backend          : Union[BuiltinBackends, BackendContract]   = _default_backend,
        singleton        : bool                                      = _default_singleton,
        encoding         : Optional[str]                             = _default_encoding,
        writable         : Optional[bool]                            = _default_writable,
        transaction      : Optional[bool]                            = _default_transaction,
        cache_locpath    : Optional[str]                             = None,
        cache_fs         : Optional[AbstractFileSystem]              = None,
        timestamp_col    : Optional[str]                             = _default_timestamp_col,
        join_level_name  : Optional[str]                             = _default_join_level_name,
        join_separator   : Optional[str]                             = _default_join_separator,
        fs               : Optional[AbstractFileSystem]              = None,
) -> Filoc:
    frontend_impl = _get_frontend(frontend)
    backend_impl  = _get_backend(backend, singleton, encoding)

    if isinstance(locpath, dict):
        # Case of composite filoc
        filoc_by_name = dict()
        for sub_filoc_name, sub_filoc in locpath.items():
            if isinstance(sub_filoc, str):
                sub_locpath = sub_filoc
                filoc_instance = filoc(
                    locpath         = sub_locpath    ,
                    frontend        = frontend       ,
                    backend         = backend        ,
                    singleton       = singleton      ,
                    writable        = writable       ,
                    transaction     = transaction    ,
                    cache_locpath   = None           ,  # Remark: currently cache is not forwarded to sub-filocs
                    cache_fs        = None           ,  # Remark: currently cache is not forwarded to sub-filocs
                    timestamp_col   = timestamp_col  ,
                    join_level_name = join_level_name,
                    join_separator  = join_separator ,
                    fs              = fs             ,
                )
            else:
                filoc_instance = sub_filoc
            filoc_by_name[sub_filoc_name] = filoc_instance

        return FilocComposite(
            filoc_by_name           = filoc_by_name,
            frontend                = frontend_impl,
            transaction             = transaction,
            join_level_name         = join_level_name,
            join_separator          = join_separator,
        )
    elif isinstance(locpath, str):
        return FilocSingle(
            locpath       = locpath,
            writable      = writable,
            transaction   = transaction,
            frontend      = frontend_impl,
            backend       = backend_impl,
            cache_locpath = cache_locpath,
            cache_fs      = cache_fs,
            timestamp_col = timestamp_col,
            fs            = fs,
        )
    else:
        raise ValueError(f'locpath must be an instance of str or dict, but is {type(locpath)}')

@overload
def filoc_json(
        locpath         : str,
        backend         : Union[BuiltinBackends, BackendContract] = _default_backend,
        singleton       : bool                                    = _default_singleton,
        writable        : Optional[bool]                          = _default_writable,
        transaction     : Optional[bool]                          = _default_transaction,
        cache_locpath   : Optional[str]                           = None,
        cache_fs        : Optional[AbstractFileSystem]            = None,
        timestamp_col   : Optional[str]                           = _default_timestamp_col,
        join_level_name : Optional[str]                           = _default_join_level_name,
        join_separator  : Optional[str]                           = _default_join_separator,
        fs              : Optional[AbstractFileSystem]            = None,
) -> FilocSingle[Dict[str, Any], List[Dict[str, Any]]]:
    """ Same as filoc(), but with typed return value to improve IDE support """
    ...


@overload
def filoc_json(
        locpath         : Mapping[str, Union[str, Filoc]],
        backend         : Union[BuiltinBackends, BackendContract] = _default_backend,
        singleton       : bool                                    = _default_singleton,
        writable        : Optional[bool]                          = _default_writable,
        transaction     : Optional[bool]                          = _default_transaction,
        cache_locpath   : Optional[str]                           = None,
        cache_fs        : Optional[AbstractFileSystem]            = None,
        timestamp_col   : Optional[str]                           = _default_timestamp_col,
        join_level_name : Optional[str]                           = _default_join_level_name,
        join_separator  : Optional[str]                           = _default_join_separator,
        fs              : Optional[AbstractFileSystem]            = None,
) -> FilocComposite[Dict[str, Any], List[Dict[str, Any]]]:
    """ Same as filoc(), but with typed return value to improve IDE support """
    ...


def filoc_json(
        locpath         : Union[str, Mapping[str, Union[str, Filoc]]],
        backend         : Union[BuiltinBackends, BackendContract] = _default_backend,
        singleton       : bool                                    = _default_singleton,
        writable        : Optional[bool]                          = _default_writable,
        transaction     : Optional[bool]                          = _default_transaction,
        cache_locpath   : Optional[str]                           = None,
        cache_fs        : Optional[AbstractFileSystem]            = None,
        timestamp_col   : Optional[str]                           = _default_timestamp_col,
        join_level_name : Optional[str]                           = _default_join_level_name,
        join_separator  : Optional[str]                           = _default_join_separator,
        fs              : Optional[AbstractFileSystem]            = None,
) -> Union[
        FilocSingle   [Dict[str, Any], List[Dict[str, Any]]],
        FilocComposite[Dict[str, Any], List[Dict[str, Any]]]
    ]:
    """ Same as filoc(), but with typed return value to improve IDE support """
    loc = filoc(
        locpath         = locpath              ,
        frontend        = _get_frontend('json'),
        backend         = backend              ,
        singleton       = singleton            ,
        writable        = writable             ,
        transaction     = transaction          ,
        cache_locpath   = cache_locpath        ,
        cache_fs        = cache_fs             ,
        timestamp_col   = timestamp_col        ,
        join_level_name = join_level_name      ,
        join_separator  = join_separator       ,
        fs              = fs                   ,
    )
    return loc


@overload
def filoc_pandas(
        locpath         : str,
        backend         : Union[BuiltinBackends, BackendContract] = _default_backend,
        singleton       : bool                                    = _default_singleton,
        writable        : Optional[bool]                          = _default_writable,
        transaction     : Optional[bool]                          = _default_transaction,
        cache_locpath   : Optional[str]                           = None,
        cache_fs        : Optional[AbstractFileSystem]            = None,
        timestamp_col   : Optional[str]                           = _default_timestamp_col,
        join_level_name : Optional[str]                           = _default_join_level_name,
        join_separator  : Optional[str]                           = _default_join_separator,
        fs              : Optional[AbstractFileSystem]            = None,
) -> FilocSingle[Series, DataFrame]:
    """ Same as filoc(), but with typed return value to improve IDE support """
    ...


@overload
def filoc_pandas(
        locpath         : Mapping[str, Union[str, Filoc]],
        backend         : Union[BuiltinBackends, BackendContract] = _default_backend,
        singleton       : bool                                    = _default_singleton,
        writable        : Optional[bool]                          = _default_writable,
        transaction     : Optional[bool]                          = _default_transaction,
        cache_locpath   : Optional[str]                           = None,
        cache_fs        : Optional[AbstractFileSystem]            = None,
        timestamp_col   : Optional[str]                           = _default_timestamp_col,
        join_level_name : Optional[str]                           = _default_join_level_name,
        join_separator  : Optional[str]                           = _default_join_separator,
        fs              : Optional[AbstractFileSystem]            = None,
) -> FilocComposite[Series, DataFrame]:
    """ Same as filoc(), but with typed return value to improve IDE support """
    ...


def filoc_pandas(
        locpath         : Union[str, Mapping[str, Union[str, Filoc]]],
        backend         : Union[BuiltinBackends, BackendContract] = _default_backend,
        singleton       : bool                                    = _default_singleton,
        writable        : Optional[bool]                          = _default_writable,
        transaction     : Optional[bool]                          = _default_transaction,
        cache_locpath   : Optional[str]                           = None,
        cache_fs        : Optional[AbstractFileSystem]            = None,
        timestamp_col   : Optional[str]                           = _default_timestamp_col,
        join_level_name : Optional[str]                           = _default_join_level_name,
        join_separator  : Optional[str]                           = _default_join_separator,
        fs              : Optional[AbstractFileSystem]            = None,
) -> Union[
    FilocSingle[Series, DataFrame],
    FilocComposite[Series, DataFrame]
]:
    """ Same as filoc(), but with typed return value to improve IDE support """
    loc = filoc(
        locpath         = locpath                ,
        frontend        = _get_frontend('pandas'),
        backend         = backend                ,
        singleton       = singleton              ,
        writable        = writable               ,
        transaction     = transaction            ,
        cache_locpath   = cache_locpath          ,
        cache_fs        = cache_fs               ,
        timestamp_col   = timestamp_col          ,
        join_level_name = join_level_name        ,
        join_separator  = join_separator         ,
        fs              = fs                     ,
    )
    return loc



from datetime import datetime
import os
import shutil
import tempfile
import unittest
from io import UnsupportedOperation
from pathlib import Path

from filoc.filoc_io import FilocIO


def touch_file(file_path):
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    Path(file_path).touch()


# noinspection DuplicatedCode
# noinspection PyMissingOrEmptyDocstring
class TestFilocIO_format(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_str(self):
        locpath = self.test_dir + r'/{value}'
        loc = FilocIO(locpath)
        path = loc.render_path(value='abc')
        self.assertEqual(path, rf"{self.test_dir}/abc")
        touch_file(path)
        p = loc.list_paths_and_props(abc='abc')
        self.assertListEqual(p, [ (rf"{self.test_dir}/abc", {'value': 'abc'})])

    def test_int(self):
        locpath = self.test_dir + r'/{value:d}'
        loc = FilocIO(locpath)
        path = loc.render_path(value=123)
        self.assertEqual(path, rf"{self.test_dir}/123")
        touch_file(path)
        p = loc.list_paths_and_props(value=123)
        self.assertListEqual(p, [ (rf"{self.test_dir}/123", {'value': 123})])

    def test_float(self):
        locpath = self.test_dir + r'/{value:g}'
        loc = FilocIO(locpath)
        path = loc.render_path(value=3.5)
        self.assertEqual(path, rf"{self.test_dir}/3.5")
        touch_file(path)
        p = loc.list_paths_and_props(value=3.5)
        self.assertListEqual(p, [ (rf"{self.test_dir}/3.5", {'value': 3.5})])


if __name__ == '__main__':
    unittest.main()



""" Contains the default JSON frontend implementation """
import logging
from typing import Collection, Mapping

from filoc.contract import SingletonExpectedError, TContents, PropsList, TContent, FrontendContract, \
    FrontendConversionError, ReadOnlyPropsList

log = logging.getLogger('filoc')


class JsonFrontend(FrontendContract):
    """
    JSON frontend implementation.
    """

    def read_content(self, props_list: PropsList) -> TContent:
        """(see FrontendContract contract)"""

        if len(props_list) != 1:
            raise SingletonExpectedError(f'Expected singleton, got {len(props_list)} items to convert to content')
        return props_list[0]

    def read_contents(self, props_list: PropsList) -> TContents:
        """(see FrontendContract contract)"""
        return props_list

    def write_content(self, content: TContent) -> ReadOnlyPropsList:
        """(see FrontendContract contract)"""
        if isinstance(content, Mapping):
            return [content]
        else:
            raise FrontendConversionError(f'Expected instance of Mapping, got {type(content).__name__}')

    def write_contents(self, contents: TContents) -> ReadOnlyPropsList:
        """(see FrontendContract contract)"""
        if isinstance(contents, Collection):
            return contents
        else:
            raise FrontendConversionError(f'Expected instance of Collection, got {type(contents).__name__}')



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="filoc",
    version="0.0.20",
    author="jeromerg",
    author_email="jeromerg@gmx.net",
    description="eases the saving and reading of files within a structured folder tree",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jeromerg/filoc",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'fsspec',
        'parse',
        'frozendict',
        'pandas',
        'pyyaml',
    ],
    python_requires='>=3.6',
)


import json
import shutil
import tempfile
import unittest

from pandas import DataFrame

# noinspection DuplicatedCode
from filoc import filoc_pandas, filoc_pandas, FilocIO


# noinspection PyMissingOrEmptyDocstring
class TestMultiloc(unittest.TestCase):
    def setUp(self):
        self.maxDiff      = None
        self.test_dir     = tempfile.mkdtemp().replace('\\', '/')
        self.path_fmt_hyp = self.test_dir + r'/somewhere1/simid={simid:d}/epid={epid:d}/hyperparameters.json'
        self.path_fmt_res = self.test_dir + r'/somewhere1/epid={epid:d}/simid={simid:d}/result.json'
        self.hyp_loc      = filoc_pandas(self.path_fmt_hyp, writable=True)
        self.res_loc      = filoc_pandas(self.path_fmt_res, writable=True)

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_read_contents(self):
        pass

    def test_write_and_read_contents_dataframe(self):
        mloc = filoc_pandas({'hyp': self.hyp_loc, 'res': self.res_loc})

        # ACT 1
        mloc.write_contents(DataFrame([
            {"hyp.a": 100, "res.b": 1000, "shared.simid": 1, "shared.epid": 10},
            {"hyp.a": 200, "res.b": 2000, "shared.simid": 1, "shared.epid": 20},
            {"hyp.a": 300, "res.b": 3000, "shared.simid": 2, "shared.epid": 10},
            {"hyp.a": 400, "res.b": 4000, "shared.simid": 2, "shared.epid": 20},
        ]))

        wloc = FilocIO(self.path_fmt_hyp)
        with wloc.open({"simid": 1, "epid": 10}) as f: hyp1 = json.load(f)
        with wloc.open({"simid": 1, "epid": 20}) as f: hyp2 = json.load(f)
        with wloc.open({"simid": 2, "epid": 10}) as f: hyp3 = json.load(f)
        with wloc.open({"simid": 2, "epid": 20}) as f: hyp4 = json.load(f)

        self.assertEqual('{"a": 100}', json.dumps(hyp1, sort_keys=True))
        self.assertEqual('{"a": 200}', json.dumps(hyp2, sort_keys=True))
        self.assertEqual('{"a": 300}', json.dumps(hyp3, sort_keys=True))
        self.assertEqual('{"a": 400}', json.dumps(hyp4, sort_keys=True))

        wloc = FilocIO(self.path_fmt_res)
        with wloc.open({"simid": 1, "epid": 10}) as f: hyp1 = json.load(f)
        with wloc.open({"simid": 1, "epid": 20}) as f: hyp2 = json.load(f)
        with wloc.open({"simid": 2, "epid": 10}) as f: hyp3 = json.load(f)
        with wloc.open({"simid": 2, "epid": 20}) as f: hyp4 = json.load(f)

        self.assertEqual('{"b": 1000}', json.dumps(hyp1, sort_keys=True))
        self.assertEqual('{"b": 2000}', json.dumps(hyp2, sort_keys=True))
        self.assertEqual('{"b": 3000}', json.dumps(hyp3, sort_keys=True))
        self.assertEqual('{"b": 4000}', json.dumps(hyp4, sort_keys=True))

        # ACT 2
        p = mloc.read_contents(epid=10)
        self.assertEqual(len(p), 2)

        self.assertEqual(
            '[{"hyp.a": 100, "res.b": 1000, "shared.epid": 10, "shared.simid": 1}, {"hyp.a": 300, "res.b": 3000, "shared.epid": 10, "shared.simid": 2}]',
            json.dumps(p.to_dict(orient='records'), sort_keys=True))


if __name__ == '__main__':
    unittest.main()



""" Filoc default YAML backend implementation """
import os
from typing import Dict, Any

import yaml
from fsspec import AbstractFileSystem

from filoc.contract import PropsList, BackendContract, Constraints, Props
from filoc.utils import filter_and_coerce_loaded_file_content, coerce_file_content_to_write


class YamlBackend(BackendContract):
    """
    filoc backend used to read data from YAML files and write into them. This implementation is used when you call the filoc factory with the ``backend`` argument set to ``'yaml'``. Example:
    
    .. code-block:: python

        loc = filoc('/my/locpath/{id}/data.yaml', backend='yaml')
    """
    def __init__(self, is_singleton, encoding) -> None:
        super().__init__()
        self.is_singleton = is_singleton
        self.encoding     = encoding

    def read(self, fs: AbstractFileSystem, path: str, path_props : Props, constraints: Constraints) -> PropsList:
        """(see BackendContract contract) """
        with fs.open(path, encoding=self.encoding) as f:
            return filter_and_coerce_loaded_file_content(path, yaml.safe_load(f), path_props, constraints, self.is_singleton)

    def write(self, fs: AbstractFileSystem, path: str, props_list: PropsList) -> None:
        """(see BackendContract contract)"""
        fs.makedirs(os.path.dirname(path), exist_ok=True)
        with fs.open(path, 'w', encoding=self.encoding) as f:
            return yaml.dump(coerce_file_content_to_write(path, props_list, self.is_singleton), f, indent=2)



""" Filoc default CSV backend implementation """
import os
from csv import DictWriter, DictReader
from typing import Dict, Any

from fsspec import AbstractFileSystem

from filoc.contract import PropsList, BackendContract, Constraints, Props
from filoc.utils import filter_and_coerce_loaded_file_content


# TODO: Unit tests of CSV Backend and all default backends

class CsvBackend(BackendContract):
    """
    filoc backend used to read data from CSV files and write into them. This implementation is used when you call the filoc factory with the ``backend`` argument set to ``'csv'``. Example:
    
    .. code-block:: python

        loc = filoc('/my/locpath/{id}/data.csv', backend='csv')
    """

    def __init__(self, encoding) -> None:
        """(see BackendContract contract)"""
        super().__init__()
        self.encoding = encoding

    def read(self, fs: AbstractFileSystem, path: str, path_props : Props, constraints: Constraints) -> PropsList:
        """(see BackendContract contract) """
        with fs.open(path, 'r', encoding=self.encoding) as f:
            reader = DictReader(f)
            props_list = [dict(row) for row in reader]
            return filter_and_coerce_loaded_file_content(path, props_list, path_props, constraints, False)

    def write(self, fs: AbstractFileSystem, path: str, props_list: PropsList) -> None:
        """(see BackendContract contract)"""
        fs.makedirs(os.path.dirname(path), exist_ok=True)

        # once used orderedset library, but upgrade to python 3.10 failed
        fieldnames_set  = set()
        fieldnames_list = []
        for props in props_list:
            for k in props.keys():
                if k not in fieldnames_set:
                    fieldnames_set.add(k)
                    fieldnames_list.append(k)


        with fs.open(path, 'w', encoding=self.encoding) as f:
            writer = DictWriter(f, fieldnames_list)
            writer.writeheader()
            writer.writerows(props_list)



""" Contains the default Pandas frontend implementation """
import itertools
import logging
from typing import Collection, Mapping

from pandas import DataFrame, Series

from filoc.contract import FrontendConversionError, PropsList, SingletonExpectedError, TContents, TContent, \
    FrontendContract, ReadOnlyPropsList

log = logging.getLogger('filoc')


class PandasFrontend(FrontendContract):
    """
    Pandas frontend implementation. It is the default implementation provided by the ``filoc(...)`` factory.
    """
    def read_content(self, props_list: PropsList) -> TContent:
        """(see FrontendContract contract)"""
        if len(props_list) != 1:
            raise SingletonExpectedError(f'Expected singleton, got {len(props_list)} items to convert to content')

        return Series(props_list[0], dtype=object)  # dtype object in order to preserve int type, if Series contains only int and float values (elsewhere float wins)

    def read_contents(self, props_list: PropsList) -> TContents:
        """(see FrontendContract contract)"""
        return DataFrame(props_list)

    def write_content(self, content: TContent) -> ReadOnlyPropsList:
        """(see FrontendContract contract)"""
        if isinstance(content, Series):
            return [content.to_dict()]
        elif isinstance(content, Mapping):
            return [content]
        else:
            raise FrontendConversionError(f'Expected instance of Series or Mapping, got {type(content).__name__}')

    def write_contents(self, contents: TContents) -> ReadOnlyPropsList:
        """(see FrontendContract contract)"""
        if isinstance(contents, DataFrame):
            return contents.to_dict(orient='records')
        elif isinstance(contents, Collection):
            return contents
        else:
            raise FrontendConversionError(f'Expected instance of DataFrame or Collection, got {type(contents).__name__}')
